#!/usr/bin/env python
import math
import time

import numpy as np
from loguru import logger
from hgdl.hgdl import HGDL
from scipy.optimize import differential_evolution as devo, minimize
from scipy.stats import norm
from functools import partial


##########################################################################
def find_acquisition_function_maxima(gpo, acquisition_function, *,
                                     origin=None,
                                     number_of_maxima_sought=1,
                                     input_set=None,
                                     input_set_dim=None,
                                     optimization_method="global",
                                     optimization_pop_size=20,
                                     optimization_max_iter=10,
                                     optimization_tol=1e-6,
                                     optimization_x0=None,
                                     constraints=(),
                                     cost_function=None,
                                     cost_function_parameters=None,
                                     vectorized=True,
                                     x_out=None,
                                     dask_client=None,
                                     info=False):
    bounds = None
    candidates = None
    if input_set is None:
        raise Exception("optimization set has to be provided")
    if isinstance(input_set, np.ndarray):
        bounds = input_set
    elif isinstance(input_set, list):
        candidates = input_set
    else:
        raise Exception("input_set not given in an allowed format")
    opt_obj = None
    func = partial(evaluate_acquisition_function, gpo=gpo,
                   acquisition_function=acquisition_function,
                   origin=origin, dim=input_set_dim,
                   cost_function=cost_function,
                   cost_function_parameters=cost_function_parameters, x_out=x_out)
    grad = partial(gradient, func=func)

    logger.debug("====================================")
    logger.debug(f"Finding acquisition function maxima via {optimization_method} method")
    logger.debug("tolerance: {}", optimization_tol)
    logger.debug("population size: {}", optimization_pop_size)
    logger.debug("maximum number of iterations: {}", optimization_max_iter)
    logger.debug("bounds:")
    logger.debug(bounds)
    logger.debug("cost function parameters: {}", cost_function_parameters)
    logger.debug("====================================")
    if candidates is not None:
        res = np.asarray(list(map(func, candidates))).reshape(len(candidates))
        sort_indices = np.argsort(res)
        res = res[sort_indices]
        sorted_candidates = [candidates[sort_index] for sort_index in sort_indices]
        candidates = sorted_candidates
        length = min(number_of_maxima_sought, len(candidates))
        opti, func_eval, opt_obj = np.asarray(candidates[0:length]), res[0:length], None

    elif optimization_method == "global":
        opti, func_eval = differential_evolution(
            func,
            input_set,
            tol=optimization_tol,
            x0=optimization_x0,
            popsize=optimization_pop_size,
            max_iter=optimization_max_iter,
            constraints=constraints,
            vectorized=vectorized,
            disp=info
        )
        opti = np.asarray(opti)
        func_eval = np.asarray(func_eval)

    elif optimization_method == "hgdl":
        opt_obj = HGDL(func,
                       grad,
                       bounds,
                       num_epochs=optimization_max_iter,
                       local_optimizer="L-BFGS-B",
                       constraints=constraints)

        ###optimization_max_iter, tolerance here
        if optimization_x0 is not None: optimization_x0 = optimization_x0.reshape(1, -1)
        opt_obj.optimize(dask_client=dask_client, x0=optimization_x0, tolerance=optimization_tol)
        res = opt_obj.get_final()
        opt_obj.cancel_tasks()
        res = res[0:min(len(res), number_of_maxima_sought)]
        opti = np.asarray([entry["x"] for entry in res])
        func_eval = np.asarray([entry["f(x)"] for entry in res])

    elif optimization_method == "hgdlAsync":
        opt_obj = HGDL(func,
                       grad,
                       bounds,
                       num_epochs=optimization_max_iter,
                       local_optimizer="L-BFGS-B",
                       constraints=constraints)

        ###optimization_max_iter, tolerance here
        if optimization_x0 is not None: optimization_x0 = optimization_x0.reshape(1, -1)
        opt_obj.optimize(dask_client=dask_client, x0=optimization_x0, tolerance=optimization_tol)
        opti = np.zeros((1, input_set_dim))
        func_eval = np.zeros(1)

    elif optimization_method == "local":
        if optimization_x0 is not None and np.ndim(optimization_x0) == 1:
            x0 = optimization_x0
        elif optimization_x0 is not None and np.ndim(optimization_x0) == 2:
            x0 = optimization_x0[0]
        else:
            x0 = np.random.uniform(low=bounds[:, 0], high=bounds[:, 1], size=len(bounds))
        a = minimize(
            func,
            x0,
            method="L-BFGS-B",
            jac=grad,
            bounds=bounds,
            constraints=constraints,
            tol=optimization_tol,
            callback=None,
            options={"maxiter": optimization_max_iter,
                     'disp': info}
        )
        opti = np.array([a["x"]])
        func_eval = np.array(a["fun"])
        if np.ndim(func_eval) == 0: func_eval = np.array([func_eval])
        if a["success"] is False:
            logger.warning(
                "local acquisition function optimization not successful, solution replaced with random point.")
            opti = np.array(x0)
            if opti.ndim != 2: opti = np.array([opti])
            func_eval = evaluate_acquisition_function(x0, gpo=gpo, acquisition_function=acquisition_function,
                                                      origin=origin, dim=input_set_dim, cost_function=cost_function,
                                                      cost_function_parameters=cost_function_parameters, x_out=x_out)
            if np.ndim(func_eval) != 1: func_eval = np.array([func_eval])

    else:
        raise ValueError("Invalid acquisition function optimization method given.")
    if np.ndim(func_eval) != 1:
        logger.error("f(x): ", func_eval)
        logger.error("x: ", opti)
        raise Exception(
            "The output of the acquisition function optimization dim (f) != 1 or dim(x) != 2. Please check your "
            "acquisition function. It should return a 1-d numpy array")
    return opti, -func_eval, opt_obj


############################################################
############################################################
############################################################
############################################################
def evaluate_acquisition_function(x, *, gpo=None, acquisition_function=None, origin=None, dim=None,
                                  cost_function=None, cost_function_parameters=None, x_out=None):
    ##########################################################
    ####this function evaluates a default or a user-defined acquisition function
    ##########################################################
    if isinstance(x, np.ndarray):
        if np.ndim(x) == 1:
            x = x.reshape(-1, dim)
        elif np.ndim(x) > 2:
            raise Exception("Wrong input dim in `x`.")
    elif isinstance(x, list):
        x = x
    else:
        x = [x]

    if x_out is not None and np.ndim(x_out) != 1: raise Exception(
        "x_out in evaluate_acquisition_function has to be a 1d numpy array.")

    if cost_function is not None and origin is not None:
        cost_eval = cost_function(origin, x, cost_function_parameters)
    else:
        cost_eval = 1.0
    # for user defined acquisition function
    if callable(acquisition_function):
        return -acquisition_function(x, gpo) / cost_eval
    else:
        obj_eval = evaluate_gp_acquisition_function(x, acquisition_function, gpo, x_out=x_out)
        obj_eval = -obj_eval / cost_eval
        return obj_eval


def evaluate_gp_acquisition_function(x, acquisition_function, gpo, x_out):
    ##this function will always spit out a 1d numpy array because it assumes several `x`.
    ##For certain functions, this array will only have one entry
    ##for the other the length == len(x)
    if isinstance(x, np.ndarray) and np.ndim(x) == 1: raise Exception(
        "1d array given in evaluate_gp_acquisition_function. It has to be 2d")
    if x_out is None:
        if acquisition_function == "variance":
            res = gpo.posterior_covariance(x, x_out=x_out, variance_only=True)["v(x)"]
            return res
        elif acquisition_function == "relative information entropy":
            res = -gpo.gp_relative_information_entropy(x, x_out=x_out)["RIE"]
            return np.array([res])
        elif acquisition_function == "relative information entropy set":
            res = -gpo.gp_relative_information_entropy_set(x, x_out=x_out)["RIE"]
            return res
        elif acquisition_function == "ucb":
            m = gpo.posterior_mean(x, x_out=x_out)["f(x)"]
            v = gpo.posterior_covariance(x, x_out=x_out, variance_only=True)["v(x)"]
            return m + 3.0 * np.sqrt(v)
        elif acquisition_function == "lcb":
            m = gpo.posterior_mean(x, x_out=x_out)["f(x)"]
            v = gpo.posterior_covariance(x, x_out=x_out, variance_only=True)["v(x)"]
            return -(m - 3.0 * np.sqrt(v))
        elif acquisition_function == "maximum":
            res = gpo.posterior_mean(x, x_out=x_out)["f(x)"]
            return res
        elif acquisition_function == "gradient":
            mean_grad = gpo.posterior_mean_grad(x, x_out=x_out)["df/dx"]
            std = np.sqrt(gpo.posterior_covariance(x, x_out=x_out, variance_only=True)["v(x)"])
            res = np.linalg.norm(mean_grad, axis=1) * std
            return res
        elif acquisition_function == "minimum":
            res = gpo.posterior_mean(x, x_out=x_out)["f(x)"]
            return -res
        elif acquisition_function == "probability of improvement":
            m = gpo.posterior_mean(x, x_out=x_out)["f(x)"]
            std = np.sqrt(gpo.posterior_covariance(x, x_out=x_out, variance_only=True)["v(x)"])
            last_best = np.max(gpo.y_data)
            return norm.cdf((m - last_best) / (std + 1e-9))
        elif acquisition_function == "total correlation":
            return -np.array([gpo.gp_total_correlation(x, x_out=x_out)["total correlation"]])
        elif acquisition_function == "expected improvement":
            m = gpo.posterior_mean(x, x_out=x_out)["f(x)"]
            std = np.sqrt(gpo.posterior_covariance(x, x_out=x_out, variance_only=True)["v(x)"])
            last_best = np.max(gpo.y_data)
            a = (m - last_best)
            a[a < 0.] = 0.
            gamma = a / (std + 1e-9)
            pdf = norm.pdf(gamma)
            cdf = norm.cdf(gamma)
            return std * (gamma * cdf + pdf)
        elif acquisition_function == "target probability":
            try:
                a = gpo.args["a"]
                b = gpo.args["b"]
            except:
                raise Exception("Reading the arguments for acq func `target probability` failed.")
            mean = gpo.posterior_mean(x, x_out=x_out)["f(x)"]
            cov = gpo.posterior_covariance(x, x_out=x_out)["v(x)"] + 1e-9
            result = np.zeros((len(x)))
            for i in range(len(x)):
                result[i] = 0.5 * (math.erf((b - mean[i]) / np.sqrt(2. * cov[i]))) - math.erf(
                    (a - mean[i]) / np.sqrt(2. * cov[i]))
            return result
        else:
            raise Exception("No valid acquisition function string provided.")

    else:
        if acquisition_function == "variance":
            res = gpo.posterior_covariance(x, x_out=x_out, variance_only=True)["v(x)"]
            return np.sum(res.reshape(len(x), len(x_out), order="F"), axis=1)
        elif acquisition_function == "relative information entropy":
            res = -gpo.gp_relative_information_entropy(x, x_out=x_out)["RIE"]
            return np.array([res])
        elif acquisition_function == "relative information entropy set":
            res = -gpo.gp_relative_information_entropy_set(x, x_out=x_out)["RIE"]
            return res
        elif acquisition_function == "total correlation":
            return -np.array([gpo.gp_total_correlation(x, x_out=x_out)["total correlation"]])
        elif acquisition_function == "ucb":
            m = gpo.posterior_mean(x, x_out=x_out)["f(x)"]
            av_m = np.sum(m.reshape(len(x), len(x_out), order="F"), axis=1)
            v = gpo.posterior_covariance(x, x_out=x_out, variance_only=True)["v(x)"]
            av_v = np.sum(v.reshape(len(x), len(x_out), order="F"), axis=1)
            return av_m + 3.0 * np.sqrt(av_v)
        elif acquisition_function == "lcb":
            m = gpo.posterior_mean(x, x_out=x_out)["f(x)"]
            av_m = np.sum(m.reshape(len(x), len(x_out), order="F"), axis=1)
            v = gpo.posterior_covariance(x, x_out=x_out, variance_only=True)["v(x)"]
            av_v = np.sum(v.reshape(len(x), len(x_out), order="F"), axis=1)
            return -(av_m - 3.0 * np.sqrt(av_v))
        elif acquisition_function == "expected improvement":
            m = gpo.posterior_mean(x, x_out=x_out)["f(x)"]
            m = np.sum(m.reshape(len(x), len(x_out), order="F"), axis=1)
            std = np.sqrt(gpo.posterior_covariance(x, x_out=x_out, variance_only=True)["v(x)"])
            std = np.sum(std.reshape(len(x), len(x_out), order="F"), axis=1)
            last_best = np.max(gpo.y_data)
            a = (m - last_best)
            a[a < 0.] = 0.
            gamma = a / (std + 1e-9)
            pdf = norm.pdf(gamma)
            cdf = norm.cdf(gamma)
            return std * (gamma * cdf + pdf)
        else:
            raise Exception("No valid acquisition function string provided.")


def differential_evolution(func,
                           bounds,
                           tol,
                           popsize,
                           max_iter=100,
                           x0=None,
                           constraints=(),
                           disp=False,
                           vectorized=True):
    if vectorized: updating = 'deferred'
    else: updating = 'immediate'
    res = devo(partial(acq_function_vectorization_wrapper, func=func, vectorized=vectorized), bounds, tol=tol, x0=x0,
               maxiter=max_iter, popsize=popsize, polish=False, disp=disp, constraints=constraints,
               vectorized=vectorized, updating=updating)
    return [list(res["x"])], list([res["fun"]])


def acq_function_vectorization_wrapper(x, func=None, vectorized=False):
    if vectorized is True:
        acq = func(x.T)
    else:
        acq = func(x)
    return acq


def gradient(x, func=None):
    epsilon = 1e-6
    grad = np.zeros(len(x))
    for i in range(len(x)):
        new_point = np.array(x)
        new_point[i] += epsilon
        grad[i] = (func(new_point)[0] - func(x)[0]) / epsilon
    return grad



try:
    from ._version import __version__
except (ImportError, ModuleNotFoundError) as ex:
    raise RuntimeError('Running gpcam from source code requires installation. If you would like an editable source '
                       'install, use "pip install -e ." to perform and editable installation.') from ex
from loguru import logger

from .gp_optimizer import GPOptimizer
from .gp_optimizer import fvGPOptimizer
from .autonomous_experimenter import AutonomousExperimenterGP
from .autonomous_experimenter import AutonomousExperimenterFvGP

__all__ = ['GPOptimizer', 'fvGPOptimizer','AutonomousExperimenterGP','AutonomousExperimenterFvGP']

logger.disable('gpcam')



#!/usr/bin/env python
import numpy as np
from loguru import logger
from fvgp import fvGP
from fvgp import GP
from . import surrogate_model as sm
import warnings
import sys
import random


# TODO (for gpCAM)

class GPOptimizer:
    """
    This class is an optimization extension of the :doc:`fvgp <fvgp:index>` package
    for single-task (scalar-valued) Gaussian Processes.
    Gaussian Processes can be initialized, trained, and conditioned; also
    the posterior can be evaluated and used via acquisition functions,
    and plugged into optimizers to find their optima. This class inherits all methods from
    the :py:class:`fvgp.GP` class.

    V ... number of input points

    D ... input space dimensionality

    N ... arbitrary integers (N1, N2,...)


    All posterior evaluation functions are inherited from :py:class:`fvgp.GP`.
    These include, but are not limited to:

    :py:meth:`fvgp.GP.posterior_mean`

    :py:meth:`fvgp.GP.posterior_covariance`

    :py:meth:`fvgp.GP.posterior_mean_grad`

    :py:meth:`fvgp.GP.posterior_covariance_grad`

    :py:meth:`fvgp.GP.joint_gp_prior`

    :py:meth:`fvgp.GP.joint_gp_prior_grad`

    :py:meth:`fvgp.GP.gp_entropy`

    :py:meth:`fvgp.GP.gp_entropy_grad`

    :py:meth:`fvgp.GP.gp_kl_div`

    :py:meth:`fvgp.GP.gp_mutual_information`

    :py:meth:`fvgp.GP.gp_total_correlation`

    :py:meth:`fvgp.GP.gp_relative_information_entropy`

    :py:meth:`fvgp.GP.gp_relative_information_entropy_set`

    :py:meth:`fvgp.GP.posterior_probability`

    Methods for validation:

    :py:meth:`fvgp.GP.crps`

    :py:meth:`fvgp.GP.rmse`

    :py:meth:`fvgp.GP.make_2d_x_pred`

    :py:meth:`fvgp.GP.make_1d_x_pred`

    :py:meth:`fvgp.GP.log_likelihood`


    Parameters
    ----------
    x_data : np.ndarray or list, optional
        The input point positions. Shape (V x D), where D is the :py:attr:`fvgp.GP.index_set_dim`.
        For single-task GPs, the index set dimension = input space dimension.
        For multi-task GPs, the index set dimension = input space dimension + 1.
        If dealing with non-Euclidean inputs
        x_data should be a list, not a numpy array.
        If x_data is not provided here the GP will be initiated after `tell()`.
    y_data : np.ndarray, optional
        The values of the data points. Shape (V).
        If not provided here the GP will be initiated after `tell()`.
    init_hyperparameters : np.ndarray, optional
        Vector of hyperparameters used to initiate the GP.
        The default is an array of ones with the right length for the anisotropic Matern
        kernel with automatic relevance determination (ARD). If gp2Scale is
        enabled, the default kernel changes to the anisotropic Wendland kernel.
    noise_variances : np.ndarray, optional
        An numpy array defining the uncertainties/noise in the
        `y_data` in form of a point-wise variance. Shape (V).
        Note: if no noise_variances are provided here, the gp_noise_function
        callable will be used; if the callable is not provided, the noise variances
        will be set to `abs(np.mean(y_data)) / 100.0`. If
        noise covariances are required (correlated noise), make use of the `gp_kernel_function`.
        Only provide a noise function OR `noise_variances`, not both.
    compute_device : str, optional
        One of `cpu` or `gpu`, determines how linear algebra computations are executed. The default is `cpu`.
        For "gpu", pytorch has to be installed manually.
        If gp2Scale is enabled but no kernel is provided, the choice of the `compute_device`
        will be particularly important. In that case, the default Wendland kernel will be computed on
        the cpu or the gpu which will significantly change the compute time depending on the compute
        architecture.
    gp_kernel_function : Callable, optional
        A symmetric positive definite covariance function (a kernel)
        that calculates the covariance between
        data points. It is a function of the form k(x1,x2,hyperparameters).
        The input `x1` is a N1 x D array of positions, `x2` is a N2 x D
        array of positions, the hyperparameters argument
        is a 1d array of length D+1 for the default kernel and of a different
        length for user-defined kernels.
        The default is a stationary anisotropic kernel
        (`fvgp.GP.default_kernel`) which performs automatic relevance determination (ARD).
        The output is a matrix, an N1 x N2 numpy array.
    gp_kernel_function_grad : Callable, optional
        A function that calculates the derivative of the `gp_kernel_function` with respect to the hyperparameters.
        If provided, it will be used for local training (optimization) and can speed up the calculations.
        It accepts as input `x1` (a N1 x D array of positions),
        `x2` (a N2 x D array of positions) and
        `hyperparameters` (a 1d array of length D+1 for the default kernel).
        The default is a finite difference calculation.
        If `ram_economy` is True, the function's input is x1, x2, direction (int), and hyperparameters (numpy array).
        The output is a numpy array of shape (len(hps) x N).
        If `ram_economy` is `False`, the function's input is x1, x2, and hyperparameters.
        The output is a numpy array of shape (len(hyperparameters) x N1 x N2). See `ram_economy`.
    gp_mean_function : Callable, optional
        A function that evaluates the prior mean at a set of input position. It accepts as input
        an array of positions (of shape N1 x D) and hyperparameters (a 1d array of length D+1 for the default kernel).
        The return value is a 1d array of length N1. If None is provided,
        `fvgp.GP._default_mean_function` is used, which is the average of the `y_data`.
    gp_mean_function_grad : Callable, optional
        A function that evaluates the gradient of the `gp_mean_function` at
        a set of input positions with respect to the hyperparameters.
        It accepts as input an array of positions (of size N1 x D) and hyperparameters
        (a 1d array of length D+1 for the default kernel).
        The return value is a 2d array of
        shape (len(hyperparameters) x N1). If None is provided, either
        zeros are returned since the default mean function does not depend on hyperparameters,
        or a finite-difference approximation
        is used if `gp_mean_function` is provided.
    gp_noise_function : Callable, optional
        The noise function is a callable f(x,hyperparameters) that returns a
        vector (1d np.ndarray) of length(x).
        The input `x` is a numpy array of shape (N x D). The hyperparameter array is the same
        that is communicated to mean and kernel functions.
        Only provide a noise function OR a noise variance vector, not both.
        If noise covariances are required (correlated noise), make use of the `gp_kernel_function`.
    gp_noise_function_grad : Callable, optional
        A function that evaluates the gradient of the `gp_noise_function`
        at an input position with respect to the hyperparameters.
        It accepts as input an array of positions (of size N x D) and
        hyperparameters (a 1d array of length D+1 for the default kernel).
        The return value is a 2d array of
        shape (len(hyperparameters) x N). If None is provided, either
        zeros are returned since the default noise function does not depend on
        hyperparameters, or, if `gp_noise_function` is provided but no noise function,
        a finite-difference approximation will be used.
        The same rules regarding `ram_economy` as for the kernel definition apply here.
    gp2Scale: bool, optional
        Turns on gp2Scale. This will distribute the covariance computations across multiple workers.
        This is an advanced feature for HPC GPs up to 10
        million data points. If gp2Scale is used, the default kernel is an anisotropic
        Wendland kernel which is compactly supported. There are a few
        things to consider (read on); this is an advanced option.
        If no kernel is provided, the `compute_device` option should be revisited.
        The default kernel will use the specified device to compute covariances.
        The default is False.
    gp2Scale_dask_client : dask.distributed.Client, optional
        A dask client for gp2Scale.
        On HPC architecture, this client is provided by the job script. Please have a look at the examples.
        A local client is used as the default.
    gp2Scale_batch_size : int, optional
        Matrix batch size for distributed computing in gp2Scale. The default is 10000.
    gp2Scale_linalg_mode : str, optional
        One of `Chol`, `sparseLU`, `sparseCG`, `sparseMINRES`, `sparseSolve`, `sparseCGpre`
        (incomplete LU preconditioner), or `sparseMINRESpre`. The default is None which amounts to
        an automatic determination of the mode.
    calc_inv : bool, optional
        If True, the algorithm calculates and stores the inverse of the covariance
        matrix after each training or update of the dataset or hyperparameters,
        which makes computing the posterior covariance faster (3-10 times).
        For larger problems (>2000 data points), the use of inversion should be avoided due
        to computational instability and costs. The default is
        False. Note, the training will not use the
        inverse for stability reasons. Storing the inverse is
        a good option when the dataset is not too large and the posterior covariance is heavily used.
        Caution: this option, together with `append=True` in `tell()` will mean that the inverse of
        the covariance is updated, not recomputed, which can lead to instability.
        In application where data is appended many times, it is recommended to either turn
        `calc_inv` off, or to regularly force the recomputation of the inverse via `gp_rank_n_update` in
        `update_gp_data`.
    ram_economy : bool, optional
        Only of interest if the gradient and/or Hessian of the log marginal likelihood is/are used for the training.
        If True, components of the derivative of the log marginal likelihood are
        calculated sequentially, leading to a slow-down
        but much less RAM usage. If the derivative of the kernel (and noise function) with
        respect to the hyperparameters (gp_kernel_function_grad) is
        going to be provided, it has to be tailored: for `ram_economy=True` it should be
        of the form f(x, direction, hyperparameters)
        and return a 2d numpy array of shape len(x1) x len(x2).
        If `ram_economy=False`, the function should be of the form f(x, hyperparameters)
        and return a numpy array of shape
        H x len(x1) x len(x2), where H is the number of hyperparameters.
        CAUTION: This array will be stored and is very large.
    args : any, optional
        args will be a class attribute and therefore available to kernel, noise and prior mean functions.
    cost_function : Callable, optional
        A function encoding the cost of motion through the input
        space and the cost of a measurement. Its inputs
        are an `origin` (np.ndarray of size V x D), `x`
        (np.ndarray of size V x D), and the value of `cost_func_params`;
        `origin` is the starting position, and `x` is the
        destination position. The return value is a 1d array of
        length V describing the costs as floats. The 'score' from
        acquisition_function is divided by this
        returned cost to determine the next measurement point.
        The default in no-op.
    cost_function_parameters : object, optional
        This object is transmitted to the cost function;
        it can be of any type. The default is None.
    cost_update_function : Callable, optional
        If provided this function will be used when
        :py:meth:`update_cost_function` is called.
        The function `cost_update_function` accepts as
        input costs and a parameter
        object. The default is a no-op.

    Attributes
    ----------
    x_data : np.ndarray
        Datapoint positions
    y_data : np.ndarray
        Datapoint values
    noise_variances : np.ndarray
        Datapoint observation variances
    hyperparameters : np.ndarray
        Current hyperparameters in use.
    """

    def __init__(
            self,
            x_data=None,
            y_data=None,
            init_hyperparameters=None,
            noise_variances=None,
            compute_device="cpu",
            gp_kernel_function=None,
            gp_kernel_function_grad=None,
            gp_noise_function=None,
            gp_noise_function_grad=None,
            gp_mean_function=None,
            gp_mean_function_grad=None,
            gp2Scale=False,
            gp2Scale_dask_client=None,
            gp2Scale_batch_size=10000,
            gp2Scale_linalg_mode=None,
            calc_inv=False,
            ram_economy=False,
            args=None,
            cost_function=None,
            cost_function_parameters=None,
            cost_update_function=None
    ):
        self.cost_function = cost_function
        self.cost_function_parameters = cost_function_parameters
        self.cost_update_function = cost_update_function
        self.hyperparameters = init_hyperparameters
        self.compute_device = compute_device
        self.gp_kernel_function = gp_kernel_function
        self.gp_kernel_function_grad = gp_kernel_function_grad
        self.gp_noise_function = gp_noise_function
        self.gp_noise_function_grad = gp_noise_function_grad
        self.gp_mean_function = gp_mean_function
        self.gp_mean_function_grad = gp_mean_function_grad
        self.gp2Scale = gp2Scale
        self.gp2Scale_dask_client = gp2Scale_dask_client
        self.gp2Scale_batch_size = gp2Scale_batch_size
        self.gp2Scale_linalg_mode = gp2Scale_linalg_mode
        self.calc_inv = calc_inv
        self.ram_economy = ram_economy
        self.args = args
        self.gp = None
        if x_data is not None and y_data is not None:
            self._initializeGP(x_data, y_data, noise_variances=noise_variances)


    #########################################################################################
    def __getattr__(self, attr):
        if not self.gp:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'. "
                                 f"You may need to initialize the GP.")
        elif self.gp and hasattr(self.gp, attr):
            return getattr(self.gp, attr)
        else:
            raise AttributeError(f"Attribute '{attr}' not available.")

    #########################################################################################
    def _initializeGP(self, x_data, y_data, noise_variances=None):
        """
        Function to initialize a GP object.
        If data is prided at initialization this function is NOT needed.
        It has the same parameters as the initialization of the class (except cost related).
        """
        self.gp = GP(
            x_data,
            y_data,
            init_hyperparameters=self.hyperparameters,
            noise_variances=noise_variances,
            compute_device=self.compute_device,
            gp_kernel_function=self.gp_kernel_function,
            gp_kernel_function_grad=self.gp_kernel_function_grad,
            gp_noise_function=self.gp_noise_function,
            gp_noise_function_grad=self.gp_noise_function_grad,
            gp_mean_function=self.gp_mean_function,
            gp_mean_function_grad=self.gp_mean_function_grad,
            gp2Scale=self.gp2Scale,
            gp2Scale_dask_client=self.gp2Scale_dask_client,
            gp2Scale_batch_size=self.gp2Scale_batch_size,
            gp2Scale_linalg_mode=self.gp2Scale_linalg_mode,
            calc_inv=self.calc_inv,
            ram_economy=self.ram_economy,
            args=self.args
        )
        self.index_set_dim = self.gp.index_set_dim
        self.input_space_dim = self.gp.index_set_dim

    #########################################################################################
    def get_data(self):
        """
        Function that provides access to the class attributes.

        Return
        ------
        dictionary of class attributes : dict
        """

        return {
            "input dim": self.prior.index_set_dim,
            "x data": self.x_data,
            "y data": self.y_data,
            "measurement variances": self.likelihood.V,
            "hyperparameters": self.hyperparameters,
            "cost function parameters": self.cost_function_parameters,
            "cost function": self.cost_function}

    def evaluate_acquisition_function(self, x, acquisition_function="variance", origin=None):
        """
        Function to evaluate the acquisition function.

        Parameters
        ----------
        x : np.ndarray or list
            Point positions at which the acquisition function is evaluated. np.ndarray of shape (N x D) or list.
        acquisition_function : Callable, optional
            Acquisition function to execute. Callable with inputs (x,gpcam.GPOptimizer),
            where x is a V x D array of input x position. The return value is a 1d array of length V.
            The default is `variance`.
        origin : np.ndarray, optional
            If a cost function is provided this 1d numpy array of length D is used as the origin of motion.

        Return
        ------
        Evaluation : np.ndarray
            The acquisition function evaluations at all x.
        """
        if self.cost_function and origin is None:
            warnings.warn("Warning: For the cost function to be active, an origin has to be provided.")
        if origin is not None and self.cost_function is None:
            warnings.warn("Warning: An origin is given but no cost function is defined. Cost function ignored")
        try:
            res = sm.evaluate_acquisition_function(
                x, gpo=self.gp, acquisition_function=acquisition_function, origin=origin, dim=self.input_space_dim,
                cost_function=self.cost_function, cost_function_parameters=self.cost_function_parameters)

            return -res
        except Exception as ex:
            logger.error(ex)
            logger.error("Evaluating the acquisition function was not successful.")
            raise Exception("Evaluating the acquisition function was not successful.", ex)

    def tell(self, x, y, noise_variances=None, append=False, gp_rank_n_update=None):
        """
        This function can tell() the gp_optimizer class
        the data that was collected. The data will instantly be used to update the GP data.

        Parameters
        ----------
        x : np.ndarray or list
            Point positions to be communicated to the Gaussian Process; either a np.ndarray of shape (U x D)
            or a list.
        y : np.ndarray
            Point values of shape (U) to be communicated to the Gaussian Process.
        noise_variances : np.ndarray, optional
            Point value variances of shape(U) to be communicated to the Gaussian Process.
            If not provided, the GP will 1% of the y values as variances.
        append : bool, optional
            The default is True. Indicates if existent data should be appended by or overwritten with the new data.
        gp_rank_n_update : bool , optional
            Indicates whether the GP marginal should be rank-n updated or recomputed. The default
            is `gp_rank_n_update=append`, meaning if data is only appended, the rank_n_update will
            be performed.
        """

        if gp_rank_n_update is None: gp_rank_n_update = append
        if self.gp is None:
            self._initializeGP(x, y, noise_variances=noise_variances)
        else:
            self.update_gp_data(x, y, noise_variances_new=noise_variances,
                                append=append, gp_rank_n_update=gp_rank_n_update)

    ##############################################################
    def train(
            self,
            objective_function=None,
            objective_function_gradient=None,
            objective_function_hessian=None,
            hyperparameter_bounds=None,
            init_hyperparameters=None,
            method="global",
            pop_size=20,
            tolerance=0.0001,
            max_iter=120,
            local_optimizer="L-BFGS-B",
            global_optimizer="genetic",
            constraints=(),
            dask_client=None,
            info=False):

        """
        This function finds the maximum of the log marginal likelihood and therefore trains the GP (synchronously).
        This can be done on a remote cluster/computer by specifying the method to be `hgdl` and
        providing a dask client. However, in that case `fvgp.GP.train_async()` is preferred.
        The GP prior will automatically be updated with the new hyperparameters after the training.


        Parameters
        ----------
        hyperparameter_bounds : np.ndarray, optional
            A 2d numpy array of shape (N x 2), where N is the number of hyperparameters.
            The default means inferring the bounds from the communicated dataset.
            This only works for the default kernel.
        objective_function : callable, optional
            The function that will be MINIMIZED for training the GP. The form of the function is f(hyperparameters=hps)
            and returns a scalar. This function can be used to train via non-standard user-defined objectives.
            The default is the negative log marginal likelihood.
        objective_function_gradient : callable, optional
            The gradient of the function that will be MINIMIZED for training the GP.
            The form of the function is f(hyperparameters=hps)
            and returns a vector of len(hps). This function can be used to train
            via non-standard user-defined objectives.
            The default is the gradient of the negative log marginal likelihood.
        objective_function_hessian : callable, optional
            The Hessian of the function that will be MINIMIZED for training the GP.
            The form of the function is f(hyperparameters=hps)
            and returns a matrix of shape(len(hps),len(hps)). This function can be used to train
            via non-standard user-defined objectives.
            The default is the Hessian of the negative log marginal likelihood.
        init_hyperparameters : np.ndarray, optional
            Initial hyperparameters used as starting location for all optimizers.
            The default is a random draw from a uniform distribution within the `hyperparameter_bounds`.
            If you want the training to start at previously trained hyperparameters you have to specify that
            explicitly.
        method : str or Callable, optional
            The method used to train the hyperparameters.
            The options are `global`, `local`, `hgdl`, `mcmc`, and a callable.
            The callable gets a `gp.GP` instance and has to return a 1d np.ndarray of hyperparameters.
            The default is `global` (scipy's differential evolution).
            If method = `mcmc`,
            the attribute `fvgp.GP.mcmc_info` is updated and contains convergence and distribution information.
        pop_size : int, optional
            A number of individuals used for any optimizer with a global component. Default = 20.
        tolerance : float, optional
            Used as termination criterion for local optimizers. Default = 0.0001.
        max_iter : int, optional
            Maximum number of iterations for global and local optimizers. Default = 120.
        local_optimizer : str, optional
            Defining the local optimizer. Default = `L-BFGS-B`, most `scipy.optimize.minimize`
            functions are permissible.
        global_optimizer : str, optional
            Defining the global optimizer. Only applicable to `method = hgdl`. Default = `genetic`
        constraints : tuple of object instances, optional
            Equality and inequality constraints for the optimization.
            If the optimizer is `hgdl` see `hgdl.readthedocs.io`.
            If the optimizer is a `scipy` optimizer, see the scipy documentation.
        dask_client : distributed.client.Client, optional
            A Dask Distributed Client instance for distributed training if `hgdl` is used. If None is provided, a new
            `dask.distributed.Client` instance is constructed.
        info : bool, optional
            Provides a way how to access information reports during training of the GP. The default is False.
            If other information is needed please utilize `logger` as described in the online
            documentation (separately for HGDL and fvgp if needed).


        Return
        ------
        optimized hyperparameters (only fyi, gp is already updated) : np.ndarray
        """
        self.hyperparameters = self.gp.train(
            objective_function=objective_function,
            objective_function_gradient=objective_function_gradient,
            objective_function_hessian=objective_function_hessian,
            hyperparameter_bounds=hyperparameter_bounds,
            init_hyperparameters=init_hyperparameters,
            method=method,
            pop_size=pop_size,
            tolerance=tolerance,
            max_iter=max_iter,
            local_optimizer=local_optimizer,
            global_optimizer=global_optimizer,
            constraints=constraints,
            dask_client=dask_client,
            info=info)
        return self.hyperparameters

    ##############################################################
    def train_async(
            self,
            objective_function=None,
            objective_function_gradient=None,
            objective_function_hessian=None,
            hyperparameter_bounds=None,
            init_hyperparameters=None,
            max_iter=10000,
            local_optimizer="L-BFGS-B",
            global_optimizer="genetic",
            constraints=(),
            dask_client=None):

        """
        This function finds the maximum of the log marginal likelihood and therefore trains the GP asynchronously.
        This can be done on a remote cluster/computer by
        providing a dask client. This function submits the training and returns
        an object which can be given to `fvgp.GP.update_hyperparameters()`,
        which will automatically update the GP with the new hyperparameters.


        Parameters
        ----------
        hyperparameter_bounds : np.ndarray, optional
            A 2d numpy array of shape (N x 2), where N is the number of hyperparameters.
            The default means inferring the bounds from the communicated dataset.
            This only works for the default kernel.
        objective_function : callable, optional
            The function that will be MINIMIZED for training the GP. The form of the function is f(hyperparameters=hps)
            and returns a scalar. This function can be used to train via non-standard user-defined objectives.
            The default is the negative log marginal likelihood.
        objective_function_gradient : callable, optional
            The gradient of the function that will be MINIMIZED for training the GP.
            The form of the function is f(hyperparameters=hps)
            and returns a vector of len(hps). This function can be used to train
            via non-standard user-defined objectives.
            The default is the gradient of the negative log marginal likelihood.
        objective_function_hessian : callable, optional
            The Hessian of the function that will be MINIMIZED for training the GP.
            The form of the function is f(hyperparameters=hps)
            and returns a matrix of shape(len(hps),len(hps)). This function can be used to train
            via non-standard user-defined objectives.
            The default is the Hessian of the negative log marginal likelihood.
        init_hyperparameters : np.ndarray, optional
            Initial hyperparameters used as starting location for all optimizers.
            The default is a random draw from a uniform distribution within the `hyperparameter_bounds`.
            If you want the training to start at previously trained hyperparameters you have to specify that
            explicitly.
        max_iter : int, optional
            Maximum number of iterations for global and local optimizers. Default = 120.
        local_optimizer : str, optional
            Defining the local optimizer. Default = `L-BFGS-B`, most `scipy.optimize.minimize`
            functions are permissible.
        global_optimizer : str, optional
            Defining the global optimizer. Only applicable to `method = hgdl`. Default = `genetic`
        constraints : tuple of object instances, optional
            Equality and inequality constraints for the optimization.
            If the optimizer is `hgdl` see `hgdl.readthedocs.io`.
            If the optimizer is a `scipy` optimizer, see the scipy documentation.
        dask_client : distributed.client.Client, optional
            A Dask Distributed Client instance for distributed training if `hgdl` is used. If None is provided, a new
            `dask.distributed.Client` instance is constructed.


        Return
        ------
        Optimization object that can be given to `fvgp.GP.update_hyperparameters()` to update the GP : object instance
        """

        opt_obj = self.gp.train_async(
            objective_function=objective_function,
            objective_function_gradient=objective_function_gradient,
            objective_function_hessian=objective_function_hessian,
            hyperparameter_bounds=hyperparameter_bounds,
            init_hyperparameters=init_hyperparameters,
            max_iter=max_iter,
            local_optimizer=local_optimizer,
            global_optimizer=global_optimizer,
            constraints=constraints,
            dask_client=dask_client
        )
        return opt_obj

    ##############################################################
    def stop_training(self, opt_obj):
        """
        Function to stop an asynchronous `hgdl` training.
        This leaves the :py:class:`distributed.client.Client` alive.

        Parameters
        ----------
        opt_obj : object instance
            Object created by :py:meth:`train_async()`.
        """
        self.gp.stop_training(opt_obj)

    def kill_training(self, opt_obj):
        """
        Function to kill an asynchronous training. This shuts down the associated :py:class:`distributed.client.Client`.

        Parameters
        ----------
        opt_obj : object instance
            Object created by :py:meth:`train_async()`.
        """
        self.gp.kill_training(opt_obj)

    ##############################################################
    def update_hyperparameters(self, opt_obj):
        """
        Function to update the Gaussian Process hyperparameters if an asynchronous training is running.

        Parameters
        ----------
        opt_obj : object instance
            Object created by :py:meth:`train_async()`.

        Return
        ------
        hyperparameters : np.ndarray
        """

        hps = self.gp.update_hyperparameters(opt_obj)
        self.hyperparameters = hps
        return hps

    def set_hyperparameters(self, hps):
        """
        Function to set hyperparameters.

        Parameters
        ----------
        hps : np.ndarray
            A 1-d numpy array of hyperparameters.
        """
        self.hyperparameters = hps
        self.gp.set_hyperparameters(hps)

    ##############################################################
    def ask(self,
            input_set,
            position=None,
            n=1,
            acquisition_function="variance",
            method="global",
            pop_size=20,
            max_iter=20,
            tol=1e-6,
            constraints=(),
            x0=None,
            vectorized=True,
            info=False,
            dask_client=None):
        """
        Given that the acquisition device is at `position`, this function `ask()`s for
        `n` new optimal points within a given `input_set` (given as bounds or candidates)  
        using the optimization setup `method`,
        `acquisition_function_pop_size`, `max_iter`, `tol`, `constraints`, and `x0`.
        This function can also choose the best candidate of a candidate set for Bayesian optimization
        on non-Euclidean input spaces.

        Parameters
        ----------
        input_set : np.ndarray or list
            Either a numpy array of floats of shape D x 2 describing the
            search space or a set of candidates in the form of a list. If a candidate list
            is provided, `ask()` will evaluate the acquisition function on each
            element and return a sorted array of length `n`.
            This is usually desirable for non-Euclidean inputs but can be used either way. If candidates are
            Euclidean, they should be provided as a list of 1d np.ndarray`s.
        position : np.ndarray, optional
            Current position in the input space. If a cost function is 
            provided this position will be taken into account
            to guarantee a cost-efficient new suggestion. The default is None.
        n : int, optional
            The algorithm will try to return n suggestions for
            new measurements. This is either done by method = 'hgdl', or otherwise
            by maximizing the collective information gain (default).
        acquisition_function : Callable, optional
            The acquisition function accepts as input a numpy array
            of size V x D (such that V is the number of input
            points, and D is the parameter space dimensionality) and
            a :py:class:`GPOptimizer` object. The return value is 1d array
            of length V providing 'scores' for each position,
            such that the highest scored point will be measured next.
            Built-in functions can be used by one of the following keys: 
            `ucb`,`lcb`,`maximum`,
            `minimum`, `variance`,`expected_improvement`,
            `relative information entropy`,`relative information entropy set`,
            `probability of improvement`, `gradient`,`total correlation`,`target probability`.
            If None, the default function `variance`, meaning
            :py:meth:`fvgp.GP.posterior_covariance` with variance_only = True will be used.
            The acquisition function can be callable and of the form my_func(x,gpcam.GPOptimizer)
            which will be maximized (!!!), so make sure desirable new measurement points
            will be located at maxima.
            Explanations of the acquisition functions:
            variance: simply the posterior variance
            relative information entropy: the KL divergence of the prior over predictions and the posterior
            relative information entropy set: the KL divergence of the prior
            defined over predictions and the posterior point-by-point
            ucb: upper confidence bound, posterior mean + 3. std
            lcb: lower confidence bound, -(posterior mean - 3. std)
            maximum: finds the maximum of the current posterior mean
            minimum: finds the maximum of the current posterior mean
            gradient: puts focus on high-gradient regions
            probability of improvement: as the name would suggest
            expected improvement: as the name would suggest
            total correlation: extension of mutual information to more than 2 random variables
            target probability: probability of a target; needs a dictionary
            GPOptimizer.args = {'a': lower bound, 'b': upper bound} to be defined.
        method : str, optional
            A string defining the method used to find the maximum
            of the acquisition function. Choose from `global`,
            `local`, `hgdl`. HGDL is an in-house hybrid optimizer
            that is comfortable on HPC hardware.
            The default is `global`.
        pop_size : int, optional
            An integer defining the number of individuals if `global` 
            is chosen as method. The default is 20. For
            :py:mod:`hgdl` this will be overwritten
            by the `dask_client` definition.
        max_iter : int, optional
            This number defined the number of iterations 
            before the optimizer is terminated. The default is 20.
        tol : float, optional
            Termination criterion for the local optimizer.
            The default is 1e-6.
        x0 : np.ndarray, optional
            A set of points as numpy array of shape N x D,
            used as starting location(s) for the optimization
            algorithms. The default is None.
        vectorized : bool, optional
            If your acquisition function is vectorized to return the 
            solution to an array of inquiries as an array,
            this option makes the optimization faster if method = 'global'
            is used. The default is True but will be set to
            False if method is not global.
        info : bool, optional
            Print optimization information. The default is False.
        constraints : tuple of object instances, optional
            scipy constraints instances, depending on the used optimizer.
        dask_client : distributed.client.Client, optional
            A Dask Distributed Client instance for distributed
            `acquisition_function` optimization. If None is provided,
            a new :py:class:`distributed.client.Client` instance is constructed for hgdl.

        Return
        ------
        Solution : {'x': np.array(maxima), "f(x)" : np.array(func_evals), "opt_obj" : opt_obj}
            Found maxima of the acquisition function, the associated function values and optimization object
            that, only in case of `method` = `hgdl` can be queried for solutions.
        """

        logger.debug("ask() initiated with hyperparameters: {}", self.hyperparameters)
        logger.debug("optimization method: {}", method)
        logger.debug("input set:\n{}", input_set)
        logger.debug("acq func: {}", acquisition_function)

        assert isinstance(vectorized, bool)

        #check for bounds or candidate set
        #check that bounds, if they exist, are 2d
        if isinstance(input_set, np.ndarray) and np.ndim(input_set) != 2:
            raise Exception("The input_set parameter has to be a 2d np.ndarray or a list.")
        #for user-defined acquisition functions, use "hgdl" if n>1
        if n > 1 and callable(acquisition_function):
            method = "hgdl"
        #make sure that if there are bounds and n>1, method has to be global, if not hgdl
        if isinstance(input_set, np.ndarray) and n > 1 and method != "hgdl":
            vectorized = False
            method = "global"
            new_optimization_bounds = np.vstack([input_set for i in range(n)])
            input_set = new_optimization_bounds
            if acquisition_function != "total correlation" and acquisition_function != "relative information entropy":
                acquisition_function = "total correlation"
                warnings.warn("You specified n>1 and method != 'hgdl' in ask(). The acquisition function "
                              "has therefore been changed to 'total correlation'")

        if acquisition_function == "total correlation" or acquisition_function == "relative information entropy":
            vectorized = False
        if method != "global": vectorized = False

        maxima, func_evals, opt_obj = sm.find_acquisition_function_maxima(
            self.gp,
            acquisition_function,
            origin=position,
            number_of_maxima_sought=n,
            input_set=input_set,
            input_set_dim=self.input_space_dim,
            optimization_method=method,
            optimization_pop_size=pop_size,
            optimization_max_iter=max_iter,
            optimization_tol=tol,
            cost_function=self.cost_function,
            cost_function_parameters=self.cost_function_parameters,
            optimization_x0=x0,
            constraints=constraints,
            vectorized=vectorized,
            info=info,
            dask_client=dask_client)
        if n > 1: return {'x': maxima.reshape(-1, self.input_space_dim), "f(x)": func_evals,
                          "opt_obj": opt_obj}
        return {'x': np.array(maxima), "f(x)": np.array(func_evals), "opt_obj": opt_obj}

    ##############################################################
    def update_cost_function(self, measurement_costs):
        """
        This function updates the parameters for the user-defined cost function
        It essentially calls the user-given cost_update_function which
        should return the new parameters.

        Parameters
        ----------
        measurement_costs: object
            An arbitrary object that describes the costs when moving in the parameter space.
            It can be arbitrary because the cost function using the parameters and the cost_update_function
            updating the parameters are both user-defined and this object has to be in accordance
            with those definitions.
        """

        if self.cost_function_parameters is None: warnings.warn(
            "No cost_function_parameters specified. Cost update failed.")
        if callable(self.cost_update_function):
            self.cost_function_parameters = self.cost_update_function(measurement_costs, self.cost_function_parameters)
        else:
            warnings.warn("No cost_update_function available. Cost update failed.")

    def optimize(self,
                 *,
                 func,
                 search_space,
                 hyperparameter_bounds=None,
                 train_at=(10, 20, 50, 100, 200),
                 x0=None,
                 acq_func='lcb',
                 max_iter=100,
                 callback=None,
                 break_condition=None,
                 ask_max_iter=20,
                 ask_pop_size=20,
                 method="global",
                 training_method="global",
                 training_max_iter=20
                 ):
        """
        This function is a light-weight optimization loop, using `tell()` and `ask()` repeatedly
        to optimize a given function, while retraining the GP regularly. For advanced customizations please
        use those three methods in a customized loop.

        Parameters
        ----------
        func : Callable
            The function to be optimized. The callable should be of the form def f(x),
            where `x` is an element of your search space. The return is a tuple of scalars (a,b) where
            `a` is the function evaluation scalar and `b` is the noise variance scalar.
        search_space : np.ndarray or list
            In the Euclidean case this should be a 2d np.ndarray of bounds in each direction of the input space.
            In the non-Euclidean case, this should be a list of all candidates.
        hyperparameter_bounds : np.ndarray
            Bound of the hyperparameters for the training. The default will only work for the default kernel.
            Otherwise, please specify bounds for your hyperparameters.
        train_at : tuple, optional
            The list should contain the integers that indicate the data lengths
            at which to train the GP. The default = [10,20,50,100,200].
        x0 : np.ndarray, optional
            Starting positions. Corresponding to the search space either elements of
            the candidate set in form of a list or elements of the Euclidean search space in the form of a
            2d np.ndarray.
        acq_func : Callable, optional
            Default lower-confidence bound(lcb) which means minimizing the `func`.
            The acquisition function should be formulated such that MAXIMIZING it will lead to the
            desired optimization (minimization or maximization) of `func`.
            For example `lcb` (the default) MAXIMIZES -(mean - 3.0 * standard dev) which is equivalent to minimizing
            (mean - 3.0 * standard dev) which leads to finding a minimum.
        max_iter : int, optional
            The maximum number of iterations. Default=10,000,000.
        callback : Callable, optional
            Function to be called in every iteration. Form: f(x_data, y_data)
        break_condition : Callable, optional
            Callable f(x_data, y_data) that should return `True` if run is complete, otherwise `False`.
        ask_max_iter : int, optional
            Default=20. Maximum number of iteration of the global and hybrid optimizer within `ask()`.
        ask_pop_size : int, optional
            Default=20. Population size of the global and hybrid optimizer.
        method : str, optional
            Default=`global`. Method of optimization of the acquisition function.
            One of `global, `local`, `hybrid`.
        training_method : str, optional
            Default=`global`. See :py:meth:`gpcam.GPOptimizer.train`
        training_max_iter : int, optional
            Default=20. See :py:meth:`gpcam.GPOptimizer.train`


        Return
        ------
        Full traces of function values `f(x)` and arguments `x` and the last entry: dict
            Form {'trace f(x)': self.gp.y_data,
                  'trace x': self.gp.x_data,
                  'f(x)': self.gp.y_data[-1],
                  'x': self.gp.x_data[-1]}
        """
        assert callable(func)
        assert isinstance(search_space, np.ndarray) or isinstance(search_space, list)
        assert isinstance(max_iter, int)

        if not x0:
            if isinstance(search_space, list): x0 = random.sample(search_space, 10)
            if isinstance(search_space, np.ndarray): x0 = np.random.uniform(low=search_space[:, 0],
                                                                            high=search_space[:, 1],
                                                                            size=(10, len(search_space)))
        result = list(map(func, x0))
        y, v = map(np.hstack, zip(*result))
        self.tell(x=x0, y=y, noise_variances=v, append=False)
        self.train(hyperparameter_bounds=hyperparameter_bounds)
        for i in range(max_iter):
            logger.debug("iteration {}", i)
            x_new = self.ask(search_space,
                             acquisition_function=acq_func,
                             method=method,
                             pop_size=ask_pop_size,
                             max_iter=ask_max_iter,
                             tol=1e-6,
                             constraints=(),
                             info=False)["x"]
            y_new, v_new = func(x_new)
            if callable(callback): callback(self.gp.x_data, self.gp.y_data)
            self.tell(x=x_new, y=y_new, noise_variances=v_new, append=True)
            if len(self.gp.x_data) in train_at: self.train(hyperparameter_bounds=hyperparameter_bounds,
                                                           method=training_method, max_iter=training_max_iter)
            if callable(break_condition):
                if break_condition(self.gp.x_data, self.gp.y_data): break
        return {'trace f(x)': self.gp.y_data,
                'trace x': self.gp.x_data,
                'f(x)': self.gp.y_data[-1],
                'x': self.gp.x_data[-1]}


######################################################################################
######################################################################################
######################################################################################
######################################################################################
######################################################################################
######################################################################################
######################################################################################
######################################################################################
######################################################################################
######################################################################################
######################################################################################
class fvGPOptimizer:
    """
    This class is an optimization extension of the :doc:`fvgp <fvgp:index>`
    package for multi-task (vector-valued) Gaussian Processes.
    Gaussian Processes can be initialized, trained, and conditioned; also
    the posterior can be evaluated and used via acquisition functions,
    and plugged into optimizers to find their maxima. This class inherits all methods from
    the :py:class:`fvgp.fvGP` class. Check :doc:`fvgp.readthedocs.io <fvgp:index>` for a full list of capabilities.

    V ... number of input points

    Di... input space dimensionality

    No... number of outputs

    N ... arbitrary integers (N1, N2,...)


    The main logic of :doc:`fvgp <fvgp:index>` is that any multi-task GP is just a single-task GP
    over a Cartesian product space of input and output space, as long as the kernel
    is flexible enough, so prepare to work on your kernel. This is the best
    way to give the user optimal control and power. At various instances, for example
    prior-mean function, noise function, and kernel function definitions, you will
    see that the input `x` is defined over this combined space.
    For example, if your input space is a Euclidean 2d space and your output
    is labelled [0,1], the input to the mean, kernel, and noise function might be

    x =

    [[0.2, 0.3,0],[0.9,0.6,0],

    [0.2, 0.3,1],[0.9,0.6,1]]

    This has to be understood and taken into account when customizing :doc:`fvgp <fvgp:index>` for multi-task
    use. The examples will provide deeper insights.


    All posterior evaluation functions are inherited from :py:class:`fvgp.GP`.
    Check there for a full list of capabilities. Methods for validation are also  available.
    Inherited include, but are not limited to:

    :py:meth:`fvgp.GP.posterior_mean`

    :py:meth:`fvgp.GP.posterior_covariance`

    :py:meth:`fvgp.GP.posterior_mean_grad`

    :py:meth:`fvgp.GP.posterior_covariance_grad`

    :py:meth:`fvgp.GP.joint_gp_prior`

    :py:meth:`fvgp.GP.joint_gp_prior_grad`

    :py:meth:`fvgp.GP.gp_entropy`

    :py:meth:`fvgp.GP.gp_entropy_grad`

    :py:meth:`fvgp.GP.gp_kl_div`

    :py:meth:`fvgp.GP.gp_mutual_information`

    :py:meth:`fvgp.GP.gp_total_correlation`

    :py:meth:`fvgp.GP.gp_relative_information_entropy`

    :py:meth:`fvgp.GP.gp_relative_information_entropy_set`

    :py:meth:`fvgp.GP.posterior_probability`


    Other methods:

    :py:meth:`fvgp.GP.crps`

    :py:meth:`fvgp.GP.rmse`

    :py:meth:`fvgp.GP.make_2d_x_pred`

    :py:meth:`fvgp.GP.make_1d_x_pred`

    :py:meth:`fvgp.GP.log_likelihood`


    Parameters
    ----------
    x_data : np.ndarray or list, optional
        The input point positions. Shape (V x Di), where Di is the :py:attr:`fvgp.fvGP.input_space_dim`.
        For multi-task GPs, the index set dimension = input space dimension + 1.
        If dealing with non-Euclidean inputs
        x_data should be a list, not a numpy array.
    y_data : np.ndarray
        The values of the data points. Shape (V,No).
        It is possible that not every entry in `x_data`
        has all corresponding tasks available. In that case `y_data` may have np.nan as the corresponding entries.
    init_hyperparameters : np.ndarray, optional
        Vector of hyperparameters used to initiate the GP.
        The default is an array of ones with the right length for the anisotropic Matern
        kernel with automatic relevance determination (ARD). The task direction is
        simply considered a separate dimension. If gp2Scale is
        enabled, the default kernel changes to the anisotropic Wendland kernel.
    noise_variances : np.ndarray, optional
        An numpy array or list defining the uncertainties/noise in the
        `y_data` in form of a point-wise variance. Shape (V, No).
        If `y_data` has np.nan entries, the corresponding
        `noise_variances` have to be np.nan.
        Note: if no noise_variances are provided here, the gp_noise_function
        callable will be used; if the callable is not provided, the noise variances
        will be set to `abs(np.mean(y_data)) / 100.0`. If
        noise covariances are required (correlated noise), make use of the `gp_noise_function`.
        Only provide a noise function OR `noise_variances`, not both.
    compute_device : str, optional
        One of `cpu` or `gpu`, determines how linear algebra computations are executed. The default is `cpu`.
        For "gpu", pytorch has to be installed manually.
        If gp2Scale is enabled but no kernel is provided, the choice of the `compute_device`
        will be particularly important. In that case, the default Wendland kernel will be computed on
        the cpu or the gpu which will significantly change the compute time depending on the compute
        architecture.
    gp_kernel_function : Callable, optional
        A symmetric positive definite covariance function (a kernel)
        that calculates the covariance between
        data points. It is a function of the form k(x1,x2,hyperparameters).
        The input `x1` a N1 x Di+1 array of positions, `x2` is a N2 x Di+1
        array of positions, the hyperparameters argument
        is a 1d array of length N depending on how many hyperparameters are initialized.
        The default is a stationary anisotropic kernel
        (`fvgp.GP.default_kernel`) which performs automatic relevance determination (ARD). The task
        direction is simply considered an additional dimension. This kernel should only be used for tests and in the
        simplest of cases.
        The output is a matrix, an N1 x N2 numpy array.
    gp_kernel_function_grad : Callable, optional
        A function that calculates the derivative of the `gp_kernel_function` with respect to the hyperparameters.
        If provided, it will be used for local training (optimization) and can speed up the calculations.
        It accepts as input `x1` (a N1 x Di + 1 array of positions),
        `x2` (a N2 x Di + 1 array of positions) and
        `hyperparameters` (a 1d array of length Di+2 for the default kernel).
        The default is a finite difference calculation.
        If `ram_economy` is True, the function's input is x1, x2,
        direction (int), and hyperparameters (numpy array).
        The output is a numpy array of shape (len(hps) x N).
        If `ram_economy` is `False`, the function's input is x1, x2, and hyperparameters.
        The output is a numpy array of shape (len(hyperparameters) x N1 x N2). See `ram_economy`.
    gp_mean_function : Callable, optional
        A function that evaluates the prior mean at a set of input position. It accepts as input
        an array of positions (of shape N1 x Di+1) and
         hyperparameters (a 1d array of length Di+2 for the default kernel).
        The return value is a 1d array of length N1. If None is provided,
        `fvgp.GP._default_mean_function` is used, which is the average of the `y_data`.
    gp_mean_function_grad : Callable, optional
        A function that evaluates the gradient of the `gp_mean_function` at
        a set of input positions with respect to the hyperparameters.
        It accepts as input an array of positions (of size N1 x Di+1) and hyperparameters
        (a 1d array of length Di+2 for the default kernel).
        The return value is a 2d array of
        shape (len(hyperparameters) x N1). If None is provided, either
        zeros are returned since the default mean function does not depend on hyperparameters,
        or a finite-difference approximation
        is used if `gp_mean_function` is provided.
    gp_noise_function : Callable, optional
        The noise function is a callable f(x,hyperparameters) that returns a
        vector (1d np.ndarray) of length(x).
        The input `x` is a numpy array of shape (N x Di+1). The hyperparameter array is the same
        that is communicated to mean and kernel functions.
        Only provide a noise function OR a noise variance vector, not both.
        If noise covariances are required (correlated noise), make use of the `gp_kernel_function`.
    gp_noise_function_grad : Callable, optional
        A function that evaluates the gradient of the `gp_noise_function`
        at an input position with respect to the hyperparameters.
        It accepts as input an array of positions (of size N x Di+1) and
        hyperparameters (a 1d array of length D+1 for the default kernel).
        The return value is a 2d array of
        shape (len(hyperparameters) x N). If None is provided, either
        zeros are returned since the default noise function does not depend on
        hyperparameters, or, if `gp_noise_function` is provided but no noise function,
        a finite-difference approximation will be used.
        The same rules regarding `ram_economy` as for the kernel definition apply here.
    gp2Scale: bool, optional
        Turns on gp2Scale. This will distribute the covariance computations across multiple workers.
        This is an advanced feature for HPC GPs up to 10
        million data points. If gp2Scale is used, the default kernel is an anisotropic
        Wendland kernel which is compactly supported. There are a few
        things to consider (read on); this is an advanced option.
        If no kernel is provided, the `compute_device` option should be revisited.
        The default kernel will use the specified device to compute covariances.
        The default is False.
    gp2Scale_dask_client : dask.distributed.Client, optional
        A dask client for gp2Scale.
        On HPC architecture, this client is provided by the job script. Please have a look at the examples.
        A local client is used as the default.
    gp2Scale_batch_size : int, optional
        Matrix batch size for distributed computing in gp2Scale. The default is 10000.
    gp2Scale_linalg_mode : str, optional
        One of `Chol`, `sparseLU`, `sparseCG`, `sparseMINRES`, `sparseSolve`, `sparseCGpre`
        (incomplete LU preconditioner), or `sparseMINRESpre`. The default is None which amounts to
        an automatic determination of the mode.
    calc_inv : bool, optional
        If True, the algorithm calculates and stores the inverse of the covariance
        matrix after each training or update of the dataset or hyperparameters,
        which makes computing the posterior covariance faster (3-10 times).
        For larger problems (>2000 data points), the use of inversion should be avoided due
        to computational instability and costs. The default is
        False. Note, the training will not use the
        inverse for stability reasons. Storing the inverse is
        a good option when the dataset is not too large and the posterior covariance is heavily used.
        Caution: this option, together with `append=True` in `tell()` will mean that the inverse of
        the covariance is updated, not recomputed, which can lead to instability.
        In application where data is appended many times, it is recommended to either turn
        `calc_inv` off, or to regularly force the recomputation of the inverse via `gp_rank_n_update` in
        `update_gp_data`.
    ram_economy : bool, optional
        Only of interest if the gradient and/or Hessian of the log marginal likelihood is/are used for the training.
        If True, components of the derivative of the log marginal likelihood are
        calculated sequentially, leading to a slow-down
        but much less RAM usage. If the derivative of the kernel (and noise function) with
        respect to the hyperparameters (gp_kernel_function_grad) is
        going to be provided, it has to be tailored: for `ram_economy=True` it should be
        of the form f(x, direction, hyperparameters)
        and return a 2d numpy array of shape len(x1) x len(x2).
        If `ram_economy=False`, the function should be of the form f(x, hyperparameters)
        and return a numpy array of shape
        H x len(x1) x len(x2), where H is the number of hyperparameters.
        CAUTION: This array will be stored and is very large.
    args : any, optional
        args will be a class attribute and therefore available to kernel, noise and prior mean functions.
    cost_function : Callable, optional
        A function encoding the cost of motion through the input
        space and the cost of a measurement. Its inputs
        are an `origin` (np.ndarray of size V x D), `x`
        (np.ndarray of size V x D), and the value of `cost_func_params`;
        `origin` is the starting position, and `x` is the
        destination position. The return value is a 1d array of
        length V describing the costs as floats. The 'score' from
        acquisition_function is divided by this
        returned cost to determine the next measurement point.
        The default in no-op.
    cost_function_parameters : object, optional
        This object is transmitted to the cost function;
        it can be of any type. The default is None.
    cost_update_function : Callable, optional
        If provided this function will be used when
        :py:meth:`update_cost_function` is called.
        The function `cost_update_function` accepts as
        input costs and a parameter
        object. The default is a no-op.


    Attributes
    ----------
    x_data : np.ndarray
        Datapoint positions
    y_data : np.ndarray
        Datapoint values
    fvgp_x_data : np.ndarray
        Datapoint positions as seen by fvgp
    fvgp_y_data : np.ndarray
        Datapoint values as seen by fvgp
    noise_variances : np.ndarray
        Datapoint observation variances.
    hyperparameters : np.ndarray
        Current hyperparameters in use.

    """

    def __init__(
            self,
            x_data=None,
            y_data=None,
            init_hyperparameters=None,
            noise_variances=None,
            compute_device="cpu",
            gp_kernel_function=None,
            gp_kernel_function_grad=None,
            gp_noise_function=None,
            gp_noise_function_grad=None,
            gp_mean_function=None,
            gp_mean_function_grad=None,
            gp2Scale=False,
            gp2Scale_dask_client=None,
            gp2Scale_batch_size=10000,
            gp2Scale_linalg_mode=None,
            calc_inv=False,
            ram_economy=False,
            args=None,
            cost_function=None,
            cost_function_parameters=None,
            cost_update_function=None,
    ):
        self.cost_function = cost_function
        self.cost_function_parameters = cost_function_parameters
        self.cost_update_function = cost_update_function
        self.hyperparameters = init_hyperparameters
        self.compute_device = compute_device
        self.gp_kernel_function = gp_kernel_function
        self.gp_kernel_function_grad = gp_kernel_function_grad
        self.gp_noise_function = gp_noise_function
        self.gp_noise_function_grad = gp_noise_function_grad
        self.gp_mean_function = gp_mean_function
        self.gp_mean_function_grad = gp_mean_function_grad
        self.gp2Scale = gp2Scale
        self.gp2Scale_dask_client = gp2Scale_dask_client
        self.gp2Scale_batch_size = gp2Scale_batch_size
        self.gp2Scale_linalg_mode = gp2Scale_linalg_mode
        self.calc_inv = calc_inv
        self.ram_economy = ram_economy
        self.args = args

        self.gp = None
        if x_data is not None and y_data is not None:
            self._initializefvGP(x_data, y_data, noise_variances=noise_variances)

    def __getattr__(self, attr):
        if not self.gp:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'. "
                                 f"You may need to initialize the GP.")
        elif self.gp and hasattr(self.gp, attr):
            return getattr(self.gp, attr)
        else:
            raise AttributeError(f"Attribute '{attr}' not available.")

    def _initializefvGP(self, x_data, y_data, noise_variances=None):
        """
        Function to initialize a GP object.
        If data is prided at initialization this function is NOT needed.
        It has the same parameters as the initialization of the class.
        """
        self.x_out = np.arange(y_data.shape[1])
        self.gp = fvGP(
            x_data,
            y_data,
            init_hyperparameters=self.hyperparameters,
            noise_variances=noise_variances,
            compute_device=self.compute_device,
            gp_kernel_function=self.gp_kernel_function,
            gp_kernel_function_grad=self.gp_kernel_function_grad,
            gp_noise_function=self.gp_noise_function,
            gp_noise_function_grad=self.gp_noise_function_grad,
            gp_mean_function=self.gp_mean_function,
            gp_mean_function_grad=self.gp_mean_function_grad,
            gp2Scale=self.gp2Scale,
            gp2Scale_dask_client=self.gp2Scale_dask_client,
            gp2Scale_batch_size=self.gp2Scale_batch_size,
            gp2Scale_linalg_mode=self.gp2Scale_linalg_mode,
            calc_inv=self.calc_inv,
            ram_economy=self.ram_economy,
            args=self.args,
        )
        self.index_set_dim = self.gp.index_set_dim
        self.input_space_dim = self.gp.input_space_dim

    ############################################################################
    def get_data(self):
        """
        Function that provides access to the class attributes.

        Return
        ------
        dictionary of class attributes : dict
        """

        return {
            "input dim": self.input_space_dim,
            "x data": self.x_data,
            "y data": self.y_data,
            "measurement variances": self.likelihood.V,
            "hyperparameters": self.hyperparameters,
            "cost function parameters": self.cost_function_parameters,
            "cost function": self.cost_function}

    ############################################################################
    def evaluate_acquisition_function(self, x, x_out=None, acquisition_function="variance", origin=None):
        """
        Function to evaluate the acquisition function.

        Parameters
        ----------
        x : np.ndarray or list
            Point positions at which the acquisition function is evaluated. np.ndarray of shape (N x D) or list.
        x_out : np.ndarray, optional
            Point positions in the output space.
        acquisition_function : Callable, optional
            Acquisition function to execute. Callable with inputs (x,gpcam.gp_optimizer.GPOptimizer),
            where x is a V x D array of input x_data. The return value is a 1d array of length V.
            The default is `variance`.
        origin : np.ndarray, optional
            If a cost function is provided this 1d numpy array of length D is used as the origin of motion.

        Return
        ------
        The acquisition function evaluations at all points x : np.ndarray
        """
        if x_out is None: x_out = self.x_out

        if self.cost_function and origin is None:
            warnings.warn("Warning: For the cost function to be active, an origin has to be provided.")
        x = np.array(x)
        try:
            res = sm.evaluate_acquisition_function(
                x, gpo=self.gp, acquisition_function=acquisition_function, origin=origin, dim=self.input_space_dim,
                cost_function=self.cost_function, cost_function_parameters=self.cost_function_parameters,
                x_out=x_out)
            return -res
        except Exception as ex:
            logger.error(ex)
            logger.error("Evaluating the acquisition function was not successful.")
            raise Exception("Evaluating the acquisition function was not successful.", ex)

    ############################################################################
    def tell(self, x, y, noise_variances=None, append=False, gp_rank_n_update=None):
        """
        This function can tell() the gp_optimizer class
        the data that was collected. The data will instantly be used to update the GP data.

        Parameters
        ----------
        x : np.ndarray or list
            Point positions to be communicated to the Gaussian Process; either a np.ndarray of shape (U x D)
            or a list.
        y : np.ndarray
            The values of the data points. Shape (V,No).
            It is possible that not every entry in `x_new`
            has all corresponding tasks available. In that case `y_new` may contain np.nan entries.
        noise_variances : np.ndarray, optional
            An numpy array or list defining the uncertainties/noise in the
            `y_data` in form of a point-wise variance. Shape (V, No).
            If `y_data` has np.nan entries, the corresponding
            `noise_variances` have to be np.nan.
            Note: if no noise_variances are provided here, the gp_noise_function
            callable will be used; if the callable is not provided, the noise variances
            will be set to `abs(np.mean(y_data)) / 100.0`. If
            noise covariances are required (correlated noise), make use of the gp_noise_function.
            Only provide a noise function OR `noise_variances`, not both.
        append : bool, optional
            Indication whether to append to or overwrite the existing dataset. Default = True.
            In the default case, data will be appended.
        gp_rank_n_update : bool , optional
            Indicates whether the GP marginal should be rank-n updated or recomputed. The default
            is `gp_rank_n_update=append`, meaning if data is only appended, the rank_n_update will
            be performed.
        """

        if gp_rank_n_update is None: gp_rank_n_update = append
        if self.gp is None:
            self._initializefvGP(x, y, noise_variances=noise_variances)
        else:
            self.update_gp_data(x, y, noise_variances_new=noise_variances,
                                append=append, gp_rank_n_update=gp_rank_n_update)

    ##############################################################
    def train(self,
              objective_function=None,
              objective_function_gradient=None,
              objective_function_hessian=None,
              hyperparameter_bounds=None,
              init_hyperparameters=None,
              method="global",
              pop_size=20,
              tolerance=0.0001,
              max_iter=120,
              local_optimizer="L-BFGS-B",
              global_optimizer="genetic",
              constraints=(),
              dask_client=None):

        """
        This function finds the maximum of the log marginal likelihood and therefore trains the GP (synchronously).
        This can be done on a remote cluster/computer by specifying the method to be `hgdl` and
        providing a dask client. However, in that case `fvgp.GP.train_async()` is preferred.
        The GP prior will automatically be updated with the new hyperparameters after the training.


        Parameters
        ----------
        hyperparameter_bounds : np.ndarray, optional
            A 2d numpy array of shape (N x 2), where N is the number of hyperparameters.
            The default means inferring the bounds from the communicated dataset.
            This only works for the default kernel.
        objective_function : callable, optional
            The function that will be MINIMIZED for training the GP. The form of the function is f(hyperparameters=hps)
            and returns a scalar. This function can be used to train via non-standard user-defined objectives.
            The default is the negative log marginal likelihood.
        objective_function_gradient : callable, optional
            The gradient of the function that will be MINIMIZED for training the GP.
            The form of the function is f(hyperparameters=hps)
            and returns a vector of len(hps). This function can be used to train
            via non-standard user-defined objectives.
            The default is the gradient of the negative log marginal likelihood.
        objective_function_hessian : callable, optional
            The Hessian of the function that will be MINIMIZED for training the GP.
            The form of the function is f(hyperparameters=hps)
            and returns a matrix of shape(len(hps),len(hps)). This function can be used to train
            via non-standard user-defined objectives.
            The default is the Hessian of the negative log marginal likelihood.
        init_hyperparameters : np.ndarray, optional
            Initial hyperparameters used as starting location for all optimizers.
            The default is a random draw from a uniform distribution within the `hyperparameter_bounds`.
            If you want the training to start at previously trained hyperparameters you have to specify that
            explicitly.
        method : str or Callable, optional
            The method used to train the hyperparameters.
            The options are `global`, `local`, `hgdl`, `mcmc`, and a callable.
            The callable gets a `gp.GP` instance and has to return a 1d np.ndarray of hyperparameters.
            The default is `global` (scipy's differential evolution).
            If method = `mcmc`,
            the attribute `fvgp.GP.mcmc_info` is updated and contains convergence and distribution information.
        pop_size : int, optional
            A number of individuals used for any optimizer with a global component. Default = 20.
        tolerance : float, optional
            Used as termination criterion for local optimizers. Default = 0.0001.
        max_iter : int, optional
            Maximum number of iterations for global and local optimizers. Default = 120.
        local_optimizer : str, optional
            Defining the local optimizer. Default = `L-BFGS-B`, most `scipy.optimize.minimize`
            functions are permissible.
        global_optimizer : str, optional
            Defining the global optimizer. Only applicable to `method = hgdl`. Default = `genetic`
        constraints : tuple of object instances, optional
            Equality and inequality constraints for the optimization.
            If the optimizer is `hgdl` see `hgdl.readthedocs.io`.
            If the optimizer is a `scipy` optimizer, see the scipy documentation.
        dask_client : distributed.client.Client, optional
            A Dask Distributed Client instance for distributed training if `hgdl` is used. If None is provided, a new
            `dask.distributed.Client` instance is constructed.


        Return
        ------
        optimized hyperparameters (only fyi, gp is already updated) : np.ndarray
        """
        self.hyperparameters = self.gp.train(
            objective_function=objective_function,
            objective_function_gradient=objective_function_gradient,
            objective_function_hessian=objective_function_hessian,
            hyperparameter_bounds=hyperparameter_bounds,
            init_hyperparameters=init_hyperparameters,
            method=method,
            pop_size=pop_size,
            tolerance=tolerance,
            max_iter=max_iter,
            local_optimizer=local_optimizer,
            global_optimizer=global_optimizer,
            constraints=constraints,
            dask_client=dask_client)
        return self.hyperparameters

    ##############################################################
    def train_async(self,
                    objective_function=None,
                    objective_function_gradient=None,
                    objective_function_hessian=None,
                    hyperparameter_bounds=None,
                    init_hyperparameters=None,
                    max_iter=10000,
                    local_optimizer="L-BFGS-B",
                    global_optimizer="genetic",
                    constraints=(),
                    dask_client=None
                    ):

        """
        This function finds the maximum of the log marginal likelihood and therefore trains the GP asynchronously.
        This can be done on a remote cluster/computer by
        providing a dask client. This function submits the training and returns
        an object which can be given to `fvgp.GP.update_hyperparameters()`,
        which will automatically update the GP with the new hyperparameters.


        Parameters
        ----------
        hyperparameter_bounds : np.ndarray, optional
            A 2d numpy array of shape (N x 2), where N is the number of hyperparameters.
            The default means inferring the bounds from the communicated dataset.
            This only works for the default kernel.
        objective_function : callable, optional
            The function that will be MINIMIZED for training the GP. The form of the function is f(hyperparameters=hps)
            and returns a scalar. This function can be used to train via non-standard user-defined objectives.
            The default is the negative log marginal likelihood.
        objective_function_gradient : callable, optional
            The gradient of the function that will be MINIMIZED for training the GP.
            The form of the function is f(hyperparameters=hps)
            and returns a vector of len(hps). This function can be used to train
            via non-standard user-defined objectives.
            The default is the gradient of the negative log marginal likelihood.
        objective_function_hessian : callable, optional
            The Hessian of the function that will be MINIMIZED for training the GP.
            The form of the function is f(hyperparameters=hps)
            and returns a matrix of shape(len(hps),len(hps)). This function can be used to train
            via non-standard user-defined objectives.
            The default is the Hessian of the negative log marginal likelihood.
        init_hyperparameters : np.ndarray, optional
            Initial hyperparameters used as starting location for all optimizers.
            The default is a random draw from a uniform distribution within the `hyperparameter_bounds`.
            If you want the training to start at previously trained hyperparameters you have to specify that
            explicitly.
        max_iter : int, optional
            Maximum number of iterations for global and local optimizers. Default = 120.
        local_optimizer : str, optional
            Defining the local optimizer. Default = `L-BFGS-B`, most `scipy.optimize.minimize`
            functions are permissible.
        global_optimizer : str, optional
            Defining the global optimizer. Only applicable to `method = hgdl`. Default = `genetic`
        constraints : tuple of object instances, optional
            Equality and inequality constraints for the optimization.
            If the optimizer is `hgdl` see `hgdl.readthedocs.io`.
            If the optimizer is a `scipy` optimizer, see the scipy documentation.
        dask_client : distributed.client.Client, optional
            A Dask Distributed Client instance for distributed training if `hgdl` is used. If None is provided, a new
            `dask.distributed.Client` instance is constructed.


        Return
        ------
        Optimization object that can be given to `fvgp.GP.update_hyperparameters()` to update the GP : object instance
        """

        opt_obj = self.gp.train_async(
            objective_function=objective_function,
            objective_function_gradient=objective_function_gradient,
            objective_function_hessian=objective_function_hessian,
            hyperparameter_bounds=hyperparameter_bounds,
            init_hyperparameters=init_hyperparameters,
            max_iter=max_iter,
            local_optimizer=local_optimizer,
            global_optimizer=global_optimizer,
            constraints=constraints,
            dask_client=dask_client
        )
        return opt_obj

    ##############################################################
    def stop_training(self, opt_obj):
        """
        Function to stop an asynchronous `hgdl` training.
        This leaves the :py:class:`distributed.client.Client` alive.

        Parameters
        ----------
        opt_obj : object instance
            Object created by :py:meth:`train_async()`.
        """
        self.gp.stop_training(opt_obj)

    def kill_training(self, opt_obj):
        """
        Function to kill an asynchronous training. This shuts down the associated :py:class:`distributed.client.Client`.

        Parameters
        ----------
        opt_obj : object instance
            Object created by :py:meth:`train_async()`.
        """
        self.gp.kill_training(opt_obj)

    ##############################################################
    def update_hyperparameters(self, opt_obj):
        """
        Function to update the Gaussian Process hyperparameters if an asynchronous training is running.

        Parameters
        ----------
        opt_obj : object instance
            Object created by :py:meth:`train_async()`.

        Return
        ------
        hyperparameters : np.ndarray
        """

        hps = self.gp.update_hyperparameters(opt_obj)
        self.hyperparameters = hps
        return hps

    def set_hyperparameters(self, hps):
        """
        Function to set hyperparameters.

        Parameters
        ----------
        hps : np.ndarray
            A 1-d numpy array of hyperparameters.
        """
        self.hyperparameters = hps
        self.gp.set_hyperparameters(hps)

    def ask(self,
            input_set,
            x_out=None,
            acquisition_function='variance',
            position=None,
            n=1,
            method="global",
            pop_size=20,
            max_iter=20,
            tol=1e-6,
            constraints=(),
            x0=None,
            vectorized=True,
            info=False,
            dask_client=None):

        """
        Given that the acquisition device is at `position`, this function `ask()`s for
        `n` new optimal points within a given `input_set` (given as bounds or candidates)
        using the optimization setup `method`,
        `acquisition_function_pop_size`, `max_iter`, `tol`, `constraints`, and `x0`.
        This function can also choose the best candidate of a candidate set for Bayesian optimization
        on non-Euclidean input spaces.

        Parameters
        ----------
        input_set : np.ndarray or list
            Either a numpy array of floats of shape D x 2 describing the
            search space or a set of candidates in the form of a list. If a candidate list
            is provided, `ask()` will evaluate the acquisition function on each
            element and return a sorted array of length `n`.
            This is usually desirable for non-Euclidean inputs but can be used either way. If candidates are
            Euclidean, they should be provided as a list of 1d np.ndarray`s.
        x_out : np.ndarray, optional
            The position indicating where in the output space the acquisition function should be evaluated.
            This array is of shape (No).
        position : np.ndarray, optional
            Current position in the input space. If a cost function is
            provided this position will be taken into account
            to guarantee a cost-efficient new suggestion. The default is None.
        n : int, optional
            The algorithm will try to return n suggestions for
            new measurements. This is either done by method = 'hgdl', or otherwise
            by maximizing the collective information gain (default).
        acquisition_function : Callable, optional
            The acquisition function accepts as input a numpy array
            of size V x D (such that V is the number of input
            points, and D is the parameter space dimensionality) and
            a :py:class:`GPOptimizer` object. The return value is 1d array
            of length V providing 'scores' for each position,
            such that the highest scored point will be measured next.
            Built-in functions can be used by one of the following keys:
            `variance`, `relative information entropy`,
            `relative information entropy set`, `total correlation`, `ucb`, `lcb`,
            and `expected improvement`.
            See :py:meth:`gpcam.GPOptimizer.ask` for a short explanation of these functions.
            In the multi-task case, it is highly recommended to
            deploy a user-defined acquisition function due to the intricate relationship
            of posterior distributions at different points in the output space.
            If None, the default function `variance`, meaning
            :py:meth:`fvgp.GP.posterior_covariance` with variance_only = True will be used.
            The acquisition function can be a callable function of the form my_func(x,gpcam.GPOptimizer)
            which will be maximized (!!!), so make sure desirable new measurement points
            will be located at maxima.
        method : str, optional
            A string defining the method used to find the maximum
            of the acquisition function. Choose from `global`,
            `local`, `hgdl`. HGDL is an in-house hybrid optimizer
            that is comfortable on HPC hardware.
            The default is `global`.
        pop_size : int, optional
            An integer defining the number of individuals if `global`
            is chosen as method. The default is 20. For
            :py:mod:`hgdl` this will be overwritten
            by the `dask_client` definition.
        max_iter : int, optional
            This number defined the number of iterations
            before the optimizer is terminated. The default is 20.
        tol : float, optional
            Termination criterion for the local optimizer.
            The default is 1e-6.
        x0 : np.ndarray, optional
            A set of points as numpy array of shape N x D,
            used as starting location(s) for the optimization
            algorithms. The default is None.
        vectorized : bool, optional
            If your acquisition function is vectorized to return the
            solution to an array of inquiries as an array,
            this option makes the optimization faster if method = 'global'
            is used. The default is True but will be set to
            False if method is not global.
        info : bool, optional
            Print optimization information. The default is False.
        constraints : tuple of object instances, optional
            scipy constraints instances, depending on the used optimizer.
        dask_client : distributed.client.Client, optional
            A Dask Distributed Client instance for distributed
            `acquisition_function` optimization. If None is provided,
            a new :py:class:`distributed.client.Client` instance is constructed for hgdl.

        Return
        ------
        Solution : {'x': np.array(maxima), "f(x)" : np.array(func_evals), "opt_obj" : opt_obj}
            Found maxima of the acquisition function, the associated function values and optimization object
            that, only in case of `method` = `hgdl` can be queried for solutions.
        """

        logger.debug("ask() initiated with hyperparameters: {}", self.hyperparameters)
        logger.debug("optimization method: {}", method)
        logger.debug("bounds:\n{}", input_set)
        logger.debug("acq func: {}", acquisition_function)
        if x_out is None: x_out = self.x_out

        assert isinstance(vectorized, bool)
        if isinstance(input_set, np.ndarray) and np.ndim(input_set) != 2:
            raise Exception("The input_set parameter has to be a 2d np.ndarray or a list.")
        #for user-defined acquisition functions, use "hgdl" if n>1
        if n > 1 and callable(acquisition_function):
            method = "hgdl"
        if isinstance(input_set, np.ndarray) and n > 1 and method != "hgdl":
            vectorized = False
            method = "global"
            new_optimization_bounds = np.vstack([input_set for i in range(n)])
            input_set = new_optimization_bounds
            if acquisition_function != "total correlation" and acquisition_function != "relative information entropy":
                acquisition_function = "total correlation"
                warnings.warn("You specified n>1 and method != 'hgdl' in ask(). The acquisition function "
                              "has therefore been changed to 'total correlation'.")
        if acquisition_function == "total correlation" or acquisition_function == "relative information entropy":
            vectorized = False
        if method != "global": vectorized = False

        maxima, func_evals, opt_obj = sm.find_acquisition_function_maxima(
            self.gp,
            acquisition_function,
            origin=position, number_of_maxima_sought=n,
            input_set=input_set,
            input_set_dim=self.input_space_dim,
            optimization_method=method,
            optimization_pop_size=pop_size,
            optimization_max_iter=max_iter,
            optimization_tol=tol,
            cost_function=self.cost_function,
            cost_function_parameters=self.cost_function_parameters,
            optimization_x0=x0,
            constraints=constraints,
            vectorized=vectorized,
            x_out=x_out,
            info=info,
            dask_client=dask_client)
        if n > 1: return {'x': maxima.reshape(-1, self.gp.input_space_dim), "f(x)": func_evals,
                          "opt_obj": opt_obj}
        return {'x': np.array(maxima), "f(x)": np.array(func_evals), "opt_obj": opt_obj}

    ##############################################################
    def update_cost_function(self, measurement_costs):
        """
        This function updates the parameters for the user-defined cost function.
        It essentially calls the user-given cost_update_function which
        should return the new parameters.

        Parameters
        ----------
        measurement_costs: object
            An arbitrary object that describes the costs when moving in the parameter space.
            It can be arbitrary because the cost function using the parameters and the cost_update_function
            updating the parameters are both user-defined and this object has to be in
            accordance with those definitions.
        """

        if self.cost_function_parameters is None: warnings.warn("No cost_function_parameters "
                                                                "specified. Cost update failed.")
        if callable(self.cost_update_function):
            self.cost_function_parameters = self.cost_update_function(
                measurement_costs, self.cost_function_parameters)
        else:
            warnings.warn("No cost_update_function available. Cost update failed.")

    def optimize(self,
                 *,
                 func,
                 search_space,
                 x_out=None,
                 hyperparameter_bounds=None,
                 train_at=(10, 20, 50, 100, 200),
                 x0=None,
                 acq_func='lcb',
                 max_iter=100,
                 callback=None,
                 break_condition=None,
                 ask_max_iter=20,
                 ask_pop_size=20,
                 method="global",
                 training_method="global",
                 training_max_iter=20
                 ):
        """
        This function is a light-weight optimization loop, using `tell()` and `ask()` repeatedly
        to optimize a given function, while retraining the GP regularly. For advanced customizations please
        use those three methods in a customized loop.

        Parameters
        ----------
        func : Callable
            The function to be optimized. The callable should be of the form def f(x),
            where `x` is an element of your search space. The return is a tuple of scalars (a,b) where
            `a` is a vector of function evaluations and `b` is a vector of noise variances.
        search_space : np.ndarray or list
            In the Euclidean case this should be a 2d np.ndarray of bounds in each direction of the input space.
            In the non-Euclidean case, this should be a list of all candidates.
        x_out : np.ndarray, optional
            The position indicating where in the output space the acquisition function should be evaluated.
            This array is of shape (No).
        hyperparameter_bounds : np.ndarray
            Bound of the hyperparameters for the training. The default will only work for the default kernel.
            Otherwise, please specify bounds for your hyperparameters.
        train_at : tuple, optional
            The list should contain the integers that indicate the data lengths
            at which to train the GP. The default = [10,20,50,100,200].
        x0 : np.ndarray, optional
            Starting positions. Corresponding to the search space either elements of
            the candidate set in form of a list or elements of the Euclidean search space in the form of a
            2d np.ndarray.
        acq_func : Callable, optional
            Default lower-confidence bound(lcb) which means minimizing the `func`.
            The acquisition function should be formulated such that MAXIMIZING it will lead to the
            desired optimization (minimization or maximization) of `func`.
            For example `lcb` (the default) MAXIMIZES -(mean - 3.0 * standard dev) which is equivalent to minimizing
            (mean - 3.0 * standard dev) which leads to finding a minimum.
        max_iter : int, optional
            The maximum number of iterations. Default=10,000,000.
        callback : Callable, optional
            Function to be called in every iteration. Form: f(x_data, y_data)
        break_condition : Callable, optional
            Callable f(x_data, y_data) that should return `True` if run is complete, otherwise `False`.
        ask_max_iter : int, optional
            Default=20. Maximum number of iteration of the global and hybrid optimizer within `ask()`.
        ask_pop_size : int, optional
            Default=20. Population size of the global and hybrid optimizer.
        method : str, optional
            Default=`global`. Method of optimization of the acquisition function.
            One of `global, `local`, `hybrid`.
        training_method : str, optional
            Default=`global`. See :py:meth:`gpcam.GPOptimizer.train`
        training_max_iter : int, optional
            Default=20. See :py:meth:`gpcam.GPOptimizer.train`


        Return
        ------
        Full traces of function values `f(x)` and arguments `x` and the last entry: dict
            Form {'trace f(x)': self.gp.y_data,
                  'trace x': self.gp.x_data,
                  'f(x)': self.gp.y_data[-1],
                  'x': self.gp.x_data[-1]}
        """
        assert callable(func)
        assert isinstance(search_space, np.ndarray) or isinstance(search_space, list)
        assert isinstance(max_iter, int)
        if x_out is None: x_out = self.x_out

        if not x0:
            if isinstance(search_space, list): x0 = random.sample(search_space, 10)
            if isinstance(search_space, np.ndarray): x0 = np.random.uniform(low=search_space[:, 0],
                                                                            high=search_space[:, 1],
                                                                            size=(10, len(search_space)))
        result = list(map(func, x0))
        y = np.asarray(list(map(np.hstack, zip(*result)))).reshape(-1, len(x_out))[0:len(result)]
        v = np.asarray(list(map(np.hstack, zip(*result)))).reshape(-1, len(x_out))[len(result):]
        self.tell(x=x0, y=y, noise_variances=v, append=False)
        self.train(hyperparameter_bounds=hyperparameter_bounds)
        for i in range(max_iter):
            logger.debug("iteration {}", i)
            x_new = self.ask(search_space,
                             x_out,
                             acquisition_function=acq_func,
                             method=method,
                             pop_size=ask_pop_size,
                             max_iter=ask_max_iter,
                             tol=1e-6,
                             constraints=(),
                             info=False)["x"]
            y_new, v_new = func(x_new)
            if callable(callback): callback(self.gp.x_data, self.gp.y_data)
            self.tell(x=x_new, y=y_new, noise_variances=v_new, append=True)
            if len(self.gp.x_data) in train_at: self.train(hyperparameter_bounds=hyperparameter_bounds,
                                                           method=training_method, max_iter=training_max_iter)
            if callable(break_condition):
                if break_condition(self.gp.x_data, self.gp.y_data): break
        return {'trace f(x)': self.gp.y_data,
                'trace x': self.gp.x_data,
                'f(x)': self.gp.y_data[-1],
                'x': self.gp.x_data[-1]}



import datetime
import math
import time
import uuid
import warnings

import numpy as np


class gpData:
    """
    Data Class

    dataset is a list of dictionaries
    data arrays are numpy arrays
    """

    def __init__(self, dim, parameter_bounds):
        self.dim = dim
        self.parameter_bounds = parameter_bounds
        self.dataset = []
        self.point_number = None

    ###############################################################

    def update_dataset(self, dataset):
        self.dataset = dataset
        self.point_number = len(dataset)

    # either create random data or communicate a data set (use translate2data for numpy arrays)
    def create_random_dataset(self, length):
        """
        creates random data of "length" and creates a dataset
        """
        x = self._create_random_points(length)
        self.point_number = len(x)
        self.dataset = self.arrays2data(x)

    def inject_dataset(self, dataset):
        """
        initializes a previously-collected dataset
        !!!for initialization only!!! just use the "+" operator to update the existing dataset
        """
        self.point_number = len(self.dataset)
        self.dataset = dataset

    def arrays2data(self, x, y=None, v=None, info=None):
        """
        translates numpy arrays to the data format
        """
        assert isinstance(x, np.ndarray)
        if y is not None: assert isinstance(y, np.ndarray)
        if v is not None: assert isinstance(v, np.ndarray)

        assert np.ndim(x) == 2
        if isinstance(y, np.ndarray): assert np.ndim(y) == 1
        if isinstance(v, np.ndarray): assert np.ndim(v) == 1

        data = []
        for i in range(len(x)):
            val = None
            var = None
            if y is not None: val = y[i]
            if v is not None: var = v[i]
            data.append(self.npy2dataset_entry(x[i], val, var))
            if info is not None: data[i].update(info[i])
        return data

    ###############################################################
    def npy2dataset_entry(self, x, y=None, v=None):
        """
        parameters:
        -----------
            x ... 1d numpy array
        """
        d = {"x_data": x, "y_data": y, "noise variance": v, "cost": None, "id": str(uuid.uuid4()),
             "time stamp": time.time(), "date time": datetime.datetime.now().strftime("%d/%m/%Y_%H:%M:%S"),
             "measured": False}
        return d

    ################################################################
    ########Extracting##############################################
    ################################################################
    def extract_data(self):
        x = self.extract_points_from_data()
        y = self.extract_y_data_from_data()
        v = self.extract_variances_from_data()
        t = self.extract_times_from_data()
        c = self.extract_costs_from_data()
        return x, y, v, t, c

    def extract_points_from_data(self):
        P = np.zeros((self.point_number, self.dim))
        for idx_data in range(self.point_number):
            P[idx_data] = self.dataset[idx_data]["x_data"]
        return P

    def extract_y_data_from_data(self):
        M = np.zeros(self.point_number)
        for idx_data in range(self.point_number):
            M[idx_data] = self.dataset[idx_data]["y_data"]
        return M

    def extract_variances_from_data(self):
        Variance = np.zeros(self.point_number)
        for idx_data in range(self.point_number):
            if self.dataset[idx_data]["noise variance"] is None: return None
            Variance[idx_data] = self.dataset[idx_data]["noise variance"]
        return Variance

    def extract_costs_from_data(self):
        Costs = []
        for idx in range(self.point_number):
            Costs.append(self.dataset[idx]["cost"])
        return Costs

    def extract_times_from_data(self):
        times = np.zeros(self.point_number)
        for idx_data in range(self.point_number):
            times[idx_data] = self.dataset[idx_data]["time stamp"]
        return times

    ###############################################################
    #######Creating################################################
    ###############################################################
    def _create_random_x(self):
        return np.random.uniform(low=self.parameter_bounds[:, 0],
                                 high=self.parameter_bounds[:, 1],
                                 size=self.dim)

    def _create_random_points(self, length):
        x = np.empty((length, self.dim))
        for i in range(length):
            x[i, :] = self._create_random_x()
        return x

    ###############################################################
    #########Cleaning##############################################
    ###############################################################
    def check_incoming_data(self):
        try:
            for entry in self.dataset:
                if entry["y_data"] is None:
                    raise Exception("Entry with no specified y_data in communicated list of data dictionaries")
                if entry["x_data"] is None:
                    raise Exception("Entry with no specified x_data in communicated list of data dictionaries")
                if entry["noise variance"] is None:
                    raise Exception("Entry with no specified noise variance in communicated list of data dictionaries")
        except:
            raise Exception(
                "Checking the incoming data could not be accomplished. This normally means that wrong formats were "
                "communicated")

    def clean_data_NaN(self):
        for entry in self.dataset:
            if self._nan_in_dict(entry):
                warnings.warn("CAUTION, NaN detected in data")
                self.dataset.remove(entry)
        self.point_number = len(self.dataset)

    def nan_in_dataset(self):
        for entry in self.dataset:
            if self._nan_in_dict(entry):
                return True
        return False

    def _nan_in_dict(self, dictionary):
        is_nan = False
        try:
            for key in dictionary:
                if type(dictionary[key]) is dict:
                    is_nan = self._nan_in_dict(dictionary[key])
                elif type(dictionary[key]) is float and math.isnan(dictionary[key]):
                    is_nan = True
                elif type(dictionary[key]) is np.ndarray and any(np.isnan(dictionary[key])):
                    is_nan = True
                elif type(dictionary[key]) is list and any(np.isnan(np.asarray(dictionary[key]))):
                    is_nan = True
                else:
                    is_nan = False
        except:
            pass
        return is_nan


######################################################################
######################################################################
######################################################################
######################################################################
######################################################################
class fvgpData(gpData):
    def __init__(self, dim, parameter_bounds):
        super(fvgpData, self).__init__(dim, parameter_bounds)
        #self.number_of_outputs = number_of_outputs

    def create_random_dataset(self, length):
        x = self._create_random_points(length)
        self.point_number = len(x)
        self.dataset = self.arrays2data(x)

    def inject_dataset(self, dataset):
        """
        initializes a previously-collected dataset
        !!!for initiation only!!! just use the "+" operator to update the existing dataset
        """
        self.point_number = len(self.dataset)
        self.dataset = dataset

    def arrays2data(self, x, y=None, v=None, info=None):
        """
        translates numpy arrays to the data format
        """
        assert isinstance(x, np.ndarray)
        if y is not None: assert isinstance(y, np.ndarray)
        if v is not None: assert isinstance(v, np.ndarray)

        data = []
        for i in range(len(x)):
            val = None
            var = None
            if y is not None: val = y[i]
            if v is not None: var = v[i]

            data.append(self.npy2dataset_entry(x[i], val, var))
            if info is not None: data[i].update(info[i])
        return data

    def npy2dataset_entry(self, x, y=None, v=None):
        """
        parameters:
        -----------
        x ... 1d numpy array
        """
        d = {"x_data": x, "y_data": y, "noise variances": v, "cost": None,
             "id": str(uuid.uuid4()), "time stamp": time.time(),
             "date time": datetime.datetime.now().strftime("%d/%m/%Y_%H:%M:%S"), "measured": False}
        return d

    ################################################################
    ################################################################
    ################################################################
    def extract_data(self):
        x = self.extract_points_from_data()
        y = self.extract_y_data_from_data()
        v = self.extract_variances_from_data()
        t = self.extract_times_from_data()
        c = self.extract_costs_from_data()
        return x, y, v, t, c

    def extract_y_data_from_data(self):
        Y = []
        for item in self.dataset: Y.append(item["y_data"])
        return np.asarray(Y)

    def extract_variances_from_data(self):
        Variance = []
        for item in self.dataset:
            if item["noise variances"] is None:
                warnings.warn("At least one noise variances `None` in data. All noise variances will be ignored.")
                return None
            Variance.append(item["noise variances"])
        return np.asarray(Variance)

    def check_incoming_data(self):
        try:
            for entry in self.dataset:
                if entry["y_data"] is None:
                    raise Exception("Entry with no specified y_data in communicated list of data dictionaries")
                if entry["x_data"] is None:
                    raise Exception("Entry with no specified x_data in communicated list of data dictionaries")
                if entry["noise variances"] is None:
                    raise Exception("Entry with no specified noise variances in communicated list of data dictionaries")
        except Exception as e:
            raise Exception(
                "Checking the incoming data could not be accomplished. "
                "This normally means that wrong formats were communicated: ", e)



# /usr/bin/env python
import inspect
import time
import warnings
import dask
import sys
import numpy as np
from loguru import logger
from .data import fvgpData, gpData
from .gp_optimizer import GPOptimizer, fvGPOptimizer


class AutonomousExperimenterGP:
    """
    This class executes the autonomous loop for a single-task Gaussian process.
    Use class :py:class:`gpcam.AutonomousExperimenterFvGP` for multi-task experiments.
    The AutonomousExperimenter is a convenience-driven functionality that does not allow
    as much customization as using the :py:class:`gpcam.GPOptimizer` directly. But it is a great option to
    get started.


    Parameters
    ----------
    input_space : np.ndarray
        A numpy array of floats of shape D x 2 describing the input space (bounds).
        The autonomous experimenter is only able to handle Euclidean spaces.
        Please use the :py:class:`gpcam.GPOptimizer` to deal with non-Euclidean cases.
    hyperparameters : np.ndarray, optional
        Vector of hyperparameters used by the GP initially.
        This class provides methods to train hyperparameters.
        The default is a random draw from a uniform distribution
        within hyperparameter_bounds, with a shape appropriate
        for the default kernel (D + 1), which is an anisotropic Matern
        kernel with automatic relevance determination (ARD). If gp2Scale is
        enabled, the default kernel changes to the anisotropic Wendland kernel.
    hyperparameter_bounds : np.ndarray, optional
        A 2d numpy array of shape (N x 2), where N is the number of needed hyperparameters.
        The default is None, in which case the hyperparameter_bounds are estimated from the domain size
        and the initial y_data. If the data changes significantly,
        the hyperparameters and the bounds should be changed/retrained. Initial hyperparameters and bounds
        can also be set in the train calls. The default only works for the default kernels.
    instrument_function : Callable, optional
         A function that takes data points (a list of dicts), and returns the same
         with the measurement data filled in. The function is
         expected to communicate with the instrument and perform measurements,
         populating fields of the data input. `y_data` and `noise variance` have to be filled in.
    init_dataset_size : int, optional
        If `x` and `y` are not provided and `dataset` is not provided,
        `init_dataset_size` must be provided. An initial
        dataset is constructed randomly with this length. The `instrument_function`
        is immediately called to measure values
        at these initial points.
    acquisition_function : Callable, optional
        The acquisition function accepts as input a numpy array
        of size V x D (such that V is the number of input
        points, and D is the parameter space dimensionality) and
        a `GPOptimizer` object. The return value is 1d array
        of length V providing 'scores' for each position,
        such that the highest scored point will be measured next.
        Built-in functions can be used by one of the following keys:
        `ucb`,`lcb`,`maximum`,
        `minimum`, `variance`,`expected_improvement`,
        `relative information entropy`,`relative information entropy set`,
        `probability of improvement`, `gradient`,`total correlation`,`target probability`.
        If None, the default function `variance`, meaning
        `fvgp.GP.posterior_covariance` with variance_only = True will be used.
        The acquisition function can be a callable function of the form my_func(x,gpcam.GPOptimizer)
        which will be maximized (!!!), so make sure desirable new measurement points
        will be located at maxima.
        Explanations of the acquisition functions:
        variance: simply the posterior variance
        relative information entropy: the KL divergence of the prior over predictions and the posterior
        relative information entropy set: the KL divergence of the prior
        defined over predictions and the posterior point-by-point
        ucb: upper confidence bound, posterior mean + 3. std
        lcb: lower confidence bound, -(posterior mean - 3. std)
        maximum: finds the maximum of the current posterior mean
        minimum: finds the maximum of the current posterior mean
        gradient: puts focus on high-gradient regions
        probability of improvement: as the name would suggest
        expected improvement: as the name would suggest
        total correlation: extension of mutual information to more than 2 random variables
        target probability: probability of a target; needs a dictionary
        GPOptimizer.args = {'a': lower bound, 'b': upper bound} to be defined.
    cost_function : Callable, optional
        A function encoding the cost of motion through the input space and the
        cost of a measurements. Its inputs are an
        `origin` (np.ndarray of size V x D), `x` (np.ndarray of size V x D),
        and the value of `cost_function_parameters`;
        `origin` is the starting position, and `x` is the destination position. The return value is a 1d array of
        length V describing the costs as floats. The 'score' from
        acquisition_function is divided by this returned cost to determine
        the next measurement point. If None, the default is a uniform cost of 1.
    cost_update_function : Callable, optional
        A function that updates the `cost_function_parameters` which are communicated to the `cost_function`.
        This function accepts as input
        costs (a list of cost values determined by `instrument_function`), bounds (a V x 2 numpy array) and a parameters
        object. The default is a no-op.
    cost_function_parameters : Any, optional
        An object that is communicated to the `cost_function` and `cost_update_function`. The default is `{}`.
    online : bool, optional
        The default is True. `online=True` will lead to calls to `gpOptimizer.tell(append=True)` which
        potentially saves a lot of time in the GP update. The GP is updated either with an inversion update
        or a Cholesky factor update.
    kernel_function : Callable, optional
        A symmetric positive definite covariance function (a kernel)
        that calculates the covariance between
        data points. It is a function of the form k(x1,x2,hyperparameters).
        The input `x1` is a N1 x D array of positions, `x2` is a N2 x D
        array of positions, the hyperparameters argument
        is a 1d array of length D+1 for the default kernel and of a different
        length for user-defined kernels.
        The default is a stationary anisotropic kernel
        (`fvgp.GP.default_kernel`) which performs automatic relevance determination (ARD).
        The output is a matrix, an N1 x N2 numpy array.
    prior_mean_function : Callable, optional
        A function that evaluates the prior mean at a set of input position. It accepts as input
        an array of positions (of shape N1 x D) and hyperparameters (a 1d array of length D+1 for the default kernel).
        The return value is a 1d array of length N1. If None is provided,
        `fvgp.GP._default_mean_function` is used, which is the average of the `y_data`.
    noise_function : Callable optional
        The noise function is a callable f(x,hyperparameters) that returns a
        vector (1d np.ndarray) of length(x).
        The input `x` is a numpy array of shape (N x D). The hyperparameter array is the same
        that is communicated to mean and kernel functions.
        Only provide a noise function OR a noise variance vector, not both.
    run_every_iteration : Callable, optional
        A function that is run at every iteration. It accepts as input a
        `gpcam.AutonomousExperimenterGP` instance. The default is a no-op.
    x_data : np.ndarray, optional
        Initial data point positions.
    y_data : np.ndarray, optional
        Initial data point values.
    noise_variances : np.ndarray, optional
        Initial data point observation variances.
    dataset : string, optional
        A filename of a gpcam-generated file that is used to initialize a new instance.
    communicate_full_dataset : bool, optional
        If True, the full dataset will be communicated to the `instrument_function`
        on each iteration. If False, only the
        newly suggested data points will be communicated. The default is False.
    compute_device : str, optional
        One of `cpu` or `gpu`, determines how linear algebra computations are executed. The default is `cpu`.
    calc_inv : bool, optional
        If True, the algorithm calculates and stores the inverse of the covariance
        matrix after each training or update of the dataset or hyperparameters,
        which makes computing the posterior covariance faster (3-10 times).
        For larger problems (>2000 data points), the use of inversion should be avoided due
        to computational instability and costs. The default is
        False. Note, the training will not use the
        inverse for stability reasons. Storing the inverse is
        a good option when the dataset is not too large and the posterior covariance is heavily used.
        Caution: this option, together with `append=True` in `tell()` will mean that the inverse of
        the covariance is updated, not recomputed, which can lead to instability.
        In application where data is appended many times, it is recommended to either turn
        `calc_inv` off, or to regularly force the recomputation of the inverse via `gp_rank_n_update` in
        `update_gp_data`.
    training_dask_client : distributed.client.Client, optional
        A Dask Distributed Client instance for distributed training. If None is provided, a new
        `dask.distributed.Client` instance is constructed.
    acq_func_opt_dask_client : distributed.client.Client, optional
        A Dask Distributed Client instance for distributed `acquisition_function`
        optimization. If None is provided, a new
        `dask.distributed.Client` instance is constructed.


    Attributes
    ----------
    x_data : np.ndarray
        Data point positions
    y_data : np.ndarray
        Data point values
    noise_variances : np.ndarray
        Data point observation variances
    data.dataset : list
        All data
    hyperparameter_bounds : np.ndarray
        A 2d array of floats of size J x 2, such that J is the length
        matching the length of `hyperparameters` defining
        the bounds for training.
    gp_optimizer : gpcam.GPOptimizer
        A GPOptimizer instance used for initializing a Gaussian process and performing optimization of the posterior.
    """

    def __init__(self,
                 input_space,
                 hyperparameters=None,
                 hyperparameter_bounds=None,
                 instrument_function=None,
                 init_dataset_size=None,
                 acquisition_function="variance",
                 cost_function=None,
                 cost_update_function=None,
                 cost_function_parameters=None,
                 online=True,
                 kernel_function=None,
                 prior_mean_function=None,
                 noise_function=None,
                 run_every_iteration=None,
                 x_data=None, y_data=None, noise_variances=None, dataset=None,
                 communicate_full_dataset=False,
                 compute_device="cpu",
                 calc_inv=False,
                 training_dask_client=None,
                 acq_func_opt_dask_client=None,
                 gp2Scale=False,
                 gp2Scale_dask_client=None,
                 gp2Scale_batch_size=10000,
                 ram_economy=True,
                 args=None
                 ):

        if cost_function_parameters is None:
            cost_function_parameters = {}

        dim = len(input_space)
        self.input_space = input_space
        self.instrument_function = instrument_function
        self.hyperparameter_bounds = hyperparameter_bounds
        self.hyperparameters = hyperparameters
        self.acquisition_function = acquisition_function
        self.run_every_iteration = run_every_iteration
        self.communicate_full_dataset = communicate_full_dataset
        self.async_train_in_progress = False
        self.training_dask_client = training_dask_client
        self.acq_func_opt_dask_client = acq_func_opt_dask_client
        self.args = args
        self.online = online
        self.costs = None
        self.acq_func_max_opt_obj = None
        self.multi_task = False
        ################################
        # getting the data ready#########
        ################################
        if init_dataset_size is None and x_data is None and dataset is None:
            raise Exception("Either provide length of initial data or an initial dataset")
        self.data = gpData(dim, input_space)
        if x_data is None and dataset is None:
            self.data.create_random_dataset(init_dataset_size)
            if instrument_function is None: raise Exception("You need to provide an instrument function.")
            self.data.update_dataset(self.instrument_function(self.data.dataset))
        elif dataset is not None:
            self.data.inject_dataset(list(np.load(dataset, allow_pickle=True)))
            hyperparameters = self.data.dataset[-1]["hyperparameters"]
        elif x_data is not None and y_data is not None and noise_variances is not None:
            self.data.update_dataset(self.data.arrays2data(x_data, y=y_data, v=noise_variances))
        elif x_data is not None and y_data is None:
            if instrument_function is None: raise Exception("You need to provide an instrument function.")
            self.data.update_dataset(self.instrument_function(self.data.arrays2data(x_data,
                                                                                    y=y_data, v=noise_variances)))
        else:
            raise Exception("No viable option for data given!")
        self.data.check_incoming_data()
        if self.data.nan_in_dataset(): self.data.clean_data_NaN()
        self.x_data, self.y_data, self.noise_variances, self.times, self.cost = self.data.extract_data()

        self.init_dataset_size = len(self.x_data)
        ######################
        ######################
        ######################
        self.opt_obj = None
        self.gp_optimizer = GPOptimizer(x_data=self.x_data,
                                        y_data=self.y_data,
                                        init_hyperparameters=hyperparameters,
                                        noise_variances=self.noise_variances,
                                        compute_device=compute_device,
                                        gp_kernel_function=kernel_function,
                                        gp_kernel_function_grad=None,
                                        gp_noise_function=noise_function,
                                        gp_noise_function_grad=None,
                                        gp_mean_function=prior_mean_function,
                                        gp_mean_function_grad=None,
                                        gp2Scale=gp2Scale,
                                        gp2Scale_dask_client=gp2Scale_dask_client,
                                        gp2Scale_batch_size=gp2Scale_batch_size,
                                        calc_inv=calc_inv,
                                        ram_economy=ram_economy,
                                        args=args,
                                        cost_function=cost_function,
                                        cost_function_parameters=cost_function_parameters,
                                        cost_update_function=cost_update_function)

        # init costs
        logger.debug(inspect.cleandoc("""#
        ##################################################################################
        Autonomous Experimenter initialization successfully concluded
        now train(...) or train_async(...), and then go(...)
        ##################################################################################"""))

    ###################################################################################
    def train(self, init_hyperparameters=None, pop_size=10, tol=0.0001, max_iter=20, method="global", constraints=()):
        """
        This function finds the maximum of the log marginal likelihood and therefore trains the GP (synchronously).
        The `go()` command will use this function if so instructed.
        The GP prior will automatically be updated with the new hyperparameters after the training.

        Parameters
        ----------
        init_hyperparameters : np.ndarray, optional
            Initial hyperparameters used as starting location for all optimizers with local component.
            The default is a random draw from a uniform distribution within the bounds.
        pop_size : int, optional
            A number of individuals used for any optimizer with a global component. Default = 20.
        tol : float, optional
            Used as termination criterion for local optimizers. Default = 0.0001.
        max_iter : int, optional
            Maximum number of iterations for global and local optimizers. Default = 120.
        method : str, optional
            Method to be used for the training. Default is `global` which means
            a differential evolution algorithm is run with the specified parameters.
            The options are `global` or `local`, or `mcmc`.
        constraints : tuple of object instances, optional
            Equality and inequality constraints for the optimization.
            If the optimizer is `hgdl` see `hgdl.readthedocs.io`.
            If the optimizer is a scipy optimizer, see the scipy documentation.
        """

        self.gp_optimizer.train(
            hyperparameter_bounds=self.hyperparameter_bounds,
            init_hyperparameters=init_hyperparameters,
            method=method,
            pop_size=pop_size,
            tolerance=tol,
            max_iter=max_iter,
            constraints=constraints)

    def train_async(self, init_hyperparameters=None, max_iter=10000, local_method="L-BFGS-B", global_method="genetic",
                    constraints=()):
        """
        Function to train the Gaussian Process asynchronously using the HGDL optimizer. 
        The use is entirely optional; this function will be called
        as part of the go() command, if so specified. This call starts a highly parallelized optimization process,
        on an architecture specified by the dask.distributed.Client. The main purpose of this function is to
        allow for large-scale distributed training. 

        Parameters
        ----------
        init_hyperparameters : np.ndarray, optional
            Initial hyperparameters used as starting location for all optimizers with local component.
            The default is a random draw from a uniform distribution within the bounds.
        max_iter : int, optional
            Maximum number of iterations for the global method. Default = 10000
            It is important to remember here that the call is run asynchronously, so
            this number does not affect run time.
        local_method : str, optional
            Local method to be used inside HGDL. Many scipy.optimize.minimize methods
            can be used, or a user-defined callable. Please read the HGDL docs for more information.
            Default = `L-BFGS-B`.
        global_method : str, optional
            Local method to be used inside HGDL. Please read the HGDL docs for more information.
            Default = `genetic`.
        constraints : tuple of object instances, optional
            Equality and inequality constraints for the optimization.
            See `hgdl.readthedocs.io` for setting up constraints.
        """

        if self.training_dask_client is None: self.training_dask_client = dask.distributed.Client()

        logger.debug("AutonomousExperimenter starts async training with dask client:")
        self.opt_obj = self.gp_optimizer.train_async(
            hyperparameter_bounds=self.hyperparameter_bounds,
            init_hyperparameters=init_hyperparameters,
            max_iter=max_iter,
            local_optimizer=local_method,
            global_optimizer=global_method,
            constraints=constraints,
            dask_client=self.training_dask_client
        )
        logger.debug("The Autonomous Experimenter started an instance of the asynchronous training.")
        self.async_train_in_progress = True

    def kill_training(self):
        """
        Function to stop an asynchronous training. This leaves the `dask.distributed.Client` alive.
        """
        if self.async_train_in_progress:
            self.gp_optimizer.stop_training(self.opt_obj)
            self.async_train_in_progress = False
        else:
            logger.debug("no training to be killed")

    def kill_all_clients(self):
        """
        Function to kill both `dask.distributed.Client` instances.
        Will be called automatically at the end of go().
        """
        try:
            self.training_dask_client.close()
            self.acq_func_opt_dask_client.close()
            self.async_train_in_progress = False
        except Exception as ex:
            logger.error(str(ex))
            logger.error("Killing of the clients failed. Please do so manually before initializing a new one.")

    def update_hps(self):
        """
        Function to update the hyperparameters if an asynchronous training is running.
        Will be called during go() as specified.
        """
        if self.async_train_in_progress:
            self.gp_optimizer.update_hyperparameters(self.opt_obj)
            logger.debug("The Autonomous Experimenter updated the hyperparameters")
        else:
            logger.warning(
                "The autonomous experimenter could not find an instance of asynchronous training. Therefore no update.")
        logger.debug("hps: {}", self.gp_optimizer.hyperparameters)

    def _tell(self, x, y, v, append=True):
        if not self.multi_task: self.gp_optimizer.tell(x, y, noise_variances=v, append=append)
        else: self.gp_optimizer.tell(x, y, noise_variances=v,  append=append)

    def _ask(self,
             bounds=None,
             position=None,
             x_out=None,
             n=1,
             acquisition_function="variance",
             method="global",
             pop_size=20,
             max_iter=20,
             tol=1e-6,
             constraints=(),
             dask_client=None):

        if not self.multi_task:
            res = self.gp_optimizer.ask(
                bounds,
                position=position,
                n=n,
                acquisition_function=acquisition_function,
                method=method,
                pop_size=pop_size,
                max_iter=max_iter,
                tol=tol,
                constraints=constraints,
                dask_client=dask_client)
        else:
            res = self.gp_optimizer.ask(
                bounds,
                x_out,
                position=position,
                n=n,
                acquisition_function=acquisition_function,
                method=method,
                pop_size=pop_size,
                max_iter=max_iter,
                tol=tol,
                constraints=constraints,
                dask_client=dask_client)
        return res

    def _extract_data(self):
        x, y, v, t, c = self.data.extract_data()
        return x, y, v, t, c

    ###################################################################################
    def go(self, N=1e15, breaking_error=1e-50,
           retrain_globally_at=(20, 50, 100, 400),
           retrain_locally_at=(20, 40, 60, 80, 100, 200, 400, 1000),
           retrain_async_at=(),
           update_cost_func_at=(),
           acq_func_opt_setting=lambda number: "global",
           training_opt_max_iter=20,
           training_opt_pop_size=10,
           training_opt_tol=1e-6,
           acq_func_opt_max_iter=20,
           acq_func_opt_pop_size=20,
           acq_func_opt_tol=1e-6,
           acq_func_opt_tol_adjust=0.1,
           number_of_suggested_measurements=1,
           checkpoint_filename=None,
           constraints=(),
           break_condition_callable=lambda n: False
           ):
        """
        Function to start the autonomous-data-acquisition loop.

        Parameters
        ----------
        N : int, optional
            Run until N points are measured. The default is `1e15`.
        breaking_error : float, optional
            Run until breaking_error is achieved (or at max N). The default is `1e-50`.
        retrain_globally_at : Iterable [int], optional
            Retrains the hyperparameters at the given number of measurements using global optimization. The default is
            `[20,50,100,400,1000]`.
        retrain_locally_at : Iterable[int], optional
            Retrains the hyperparameters at the given number of measurements using local gradient-based optimization.
            The default is `[20,40,60,80,100,200,400,1000]`.
        retrain_async_at : Iterable[int], optional
            Retrains the hyperparameters at the given number of measurements using the HGDL algorithm. This training is
            asynchronous and can be run in a distributed fashion using `training_dask_client`. The default is `[]`.
        update_cost_func_at : Iterable[int], optional
            Calls the `update_cost_function` at the given number of measurements.
            Default = ()
        acq_func_opt_setting : Callable, optional
            A callable that accepts as input the iteration index and returns either `local`, `global`, `hgdl`.
            This switches between local gradient-based, global and hybrid optimization for the acquisition function.
            The default is `lambda number: "global" if number % 2 == 0 else "local"`.
        training_opt_max_iter : int, optional
            The maximum number of iterations for any training.
            The default value is 20.
        training_opt_pop_size : int, optional
            The population size used for any training with a global component (HGDL or standard global optimizers).
            The default value is 10.
        training_opt_tol : float, optional
            The optimization tolerance for all training optimization. The default is 1e-6.
        acq_func_opt_max_iter : int, optional
            The maximum number of iterations for the `acquisition_function` optimization. The default is 20.
        acq_func_opt_pop_size : int, optional
            The population size used for any `acquisition_function` optimization with a global
            component (HGDL or standard global
            optimizers). The default value is 20.
        acq_func_opt_tol : float, optional
            The optimization tolerance for all `acquisition_function` optimization.
            The default value is 1e-6
        acq_func_opt_tol_adjust : float, optional
            The `acquisition_function` optimization tolerance is adjusted at every iteration as a
            fraction of this value.
            The default value is 0.1 .
        number_of_suggested_measurements : int, optional
            The algorithm will try to return this many suggestions for new measurements. This may be limited by how many
            optima the algorithm may find. If greater than 1, then the `acquisition_function`
            optimization method is automatically
            set to use HGDL. The default is 1.
        checkpoint_filename : str, optional
            When provided, a checkpoint of all the accumulated data will be written to this file on each iteration.
        constraints : tuple, optional
            If provided, this subjects the acquisition function optimization to constraints.
            For the definition of the constraints, follow
            the structure your chosen optimizer requires.
        break_condition_callable : Callable, optional
            Autonomous loop will stop when this function returns True. The function takes as
            input a `gpcam.AutonomousExperimenterGP` instance.
        """
        # set up
        start_time = time.time()
        start_date_time = time.strftime("%Y-%m-%d_%H_%M_%S", time.localtime())
        logger.debug("Starting the autonomous loop...")
        i = 0
        n_measurements = len(self.x_data)
        # start the loop
        while n_measurements < N:
            logger.debug("----------------------------")
            logger.debug(f"iteration {i}")
            logger.debug(f"Run Time: {time.time() - start_time} seconds")
            logger.debug(f"Number of measurements {n_measurements}")
            logger.debug("current hps: {}", self.gp_optimizer.hyperparameters)

            ###################################################
            # ask() for new suggestions########################
            ###################################################
            current_position = self.x_data[-1]
            current_method = acq_func_opt_setting(self)
            if number_of_suggested_measurements > 1 and current_method != "hgdl": current_method = "global"
            if current_method == "hgdl" and self.acq_func_opt_dask_client is None:
                self.acq_func_opt_dask_client = dask.distributed.Client()

            if self.multi_task: x_out = np.arange(len(self.data.dataset[-1]["y_data"]))
            else: x_out = None

            res = self._ask(
                self.input_space,
                position=current_position,
                x_out=x_out,
                n=number_of_suggested_measurements,
                acquisition_function=self.acquisition_function,
                method=current_method,
                pop_size=acq_func_opt_pop_size,
                max_iter=acq_func_opt_max_iter,
                tol=acq_func_opt_tol,
                constraints=constraints,
                dask_client=self.acq_func_opt_dask_client)

            self.acq_func_max_opt_obj = res["opt_obj"]
            next_measurement_points = res["x"]
            func_evals = res["f(x)"]

            ###################################################
            # assess performance###############################
            ###################################################
            if self.multi_task:
                a = np.array(next_measurement_points)
                post_var = self.gp_optimizer.posterior_covariance(a)["v(x)"]
            else:
                post_var = self.gp_optimizer.posterior_covariance(next_measurement_points)["v(x)"]
            error = np.max(np.sqrt(post_var))
            if error < 0.:
                warnings.warn("Error < 0 encountered. This suggests some instability.")
                error = 1e-6

            ###################################################
            # adjust tolerances if necessary###################
            ###################################################
            if acq_func_opt_tol_adjust:
                acq_func_opt_tol = abs(func_evals[0]) * acq_func_opt_tol_adjust
                logger.debug("acquisition function optimization tolerance changed to: {}", acq_func_opt_tol)
            logger.debug("Next points to be requested: ")
            logger.debug(next_measurement_points)

            ###################################################
            # send suggestions to instrument and get results###
            ###################################################
            info = [{"hyperparameters": self.gp_optimizer.hyperparameters,
                     "posterior std": np.sqrt(post_var[j])} for j in range(len(next_measurement_points))]
            new_data = self.data.arrays2data(next_measurement_points, info=info)
            logger.debug("Sending request to instrument ...")
            if self.communicate_full_dataset:
                if self.online: raise Exception("You specified online=True but you communicated the whole dataset."
                                                "This violates the security protocol.")
                self.data.update_dataset(self.instrument_function(self.data.dataset + new_data))
                len_of_new_data_received = len(self.data.dataset)
            else:
                new_data_received = self.instrument_function(new_data)
                len_of_new_data_received = len(new_data_received)
                self.data.update_dataset(self.data.dataset + new_data_received)
            logger.debug("Data received")
            logger.debug("Checking if data is clean ...")
            self.data.check_incoming_data()
            #if self.data.nan_in_dataset(): self.data.clean_data_NaN()
            # update arrays and the gp_optimizer
            self.x_data, self.y_data, self.noise_variances, self.times, self.costs = self._extract_data()
            logger.debug("Communicating new data to the GP")

            ###################################################
            # tell() the GP about new data#####################
            ###################################################
            if self.online:
                logger.debug("ONLINE UPDATE")
                self._tell(self.x_data[-len_of_new_data_received:],
                           self.y_data[-len_of_new_data_received:],
                           self.noise_variances[-len_of_new_data_received:],
                           append=True)
            else:
                logger.debug("OFFLINE UPDATE")
                self._tell(self.x_data,
                           self.y_data,
                           self.noise_variances,
                           append=False)

            ###################################################
            # train() the GP###################################
            ###################################################
            if any(n in retrain_async_at for n in range(n_measurements, len(self.x_data))) and n_measurements < N:
                if self.training_dask_client is None: self.training_dask_client = dask.distributed.Client()
                logger.debug("    Starting a new asynchronous training after killing the current one.")
                self.kill_training()
                self.train_async(max_iter=10000)
            elif any(n in retrain_globally_at for n in range(n_measurements, len(self.x_data))) and n_measurements < N:
                self.kill_training()
                logger.debug("    Fresh optimization from scratch via global optimization")
                self.train(pop_size=training_opt_pop_size,
                           tol=training_opt_tol,
                           max_iter=training_opt_max_iter,
                           method="global")
            elif any(n in retrain_locally_at for n in range(n_measurements, len(self.x_data))) and n_measurements < N:
                self.kill_training()
                logger.debug("    Fresh optimization from scratch via local optimization")
                self.train(pop_size=training_opt_pop_size,
                           tol=training_opt_tol,
                           max_iter=training_opt_max_iter,
                           method="local")
            else:
                logger.debug("    No training in this round but I am trying to update the hyperparameters")
                self.update_hps()
            logger.debug("    Training successfully concluded")

            ###################################################
            # run a user-defined callable######################
            ###################################################
            if self.run_every_iteration is not None: self.run_every_iteration(self)

            ###################################################
            # save some data###################################
            ###################################################
            if checkpoint_filename:
                try:
                    np.save(checkpoint_filename, self.data.dataset)
                except Exception as e:
                    raise RuntimeError("Data not saved") from e

            ###################################################
            # update costs#####################################
            ###################################################
            if i in update_cost_func_at: self.gp_optimizer.update_cost_function(self.costs)

            ###################################################
            # break check######################################
            ###################################################
            logger.debug("Current error: {}", error)
            if breaking_error > error > 0.:
                logger.debug("Breaking error has been exceeded. {}", error)
                break
            if break_condition_callable(self): break
            # update iteration numbers
            i += 1
            n_measurements = len(self.x_data)

        # clean up
        logger.debug("killing the client... and then we are done")
        self.kill_all_clients()

        logger.debug(inspect.cleandoc("""#
            ====================================================
            The autonomous experiment was concluded successfully
            ===================================================="""))


###################################################################################
###################################################################################
###################################################################################
###################################################################################
###################################################################################
###################################################################################
class AutonomousExperimenterFvGP(AutonomousExperimenterGP):
    """
    Executes the autonomous loop for a multi-task Gaussian process.

    Parameters
    ----------
    input_space : np.ndarray
        A numpy array of floats of shape D x 2 describing the input space (bounds).
        The autonomous experimenter is only able to handle Euclidean spaces.
        Please use the :py:class:`gpcam.fvGPOptimizer`to deal with non-Euclidean cases.
    hyperparameters : np.ndarray, optional
        Vector of hyperparameters used by the GP initially.
        This class provides methods to train hyperparameters.
        The default is a random draw from a uniform distribution
        within hyperparameter_bounds, with a shape appropriate
        for the default kernel (D + 1), which is an anisotropic Matern
        kernel with automatic relevance determination (ARD). If gp2Scale is
        enabled, the default kernel changes to the anisotropic Wendland kernel.
    hyperparameter_bounds : np.ndarray, optional
        A 2d numpy array of shape (N x 2), where N is the number of needed hyperparameters.
        The default is None, in which case the hyperparameter_bounds are estimated from the domain size
        and the initial y_data. If the data changes significantly,
        the hyperparameters and the bounds should be changed/retrained. Initial hyperparameters and bounds
        can also be set in the train calls. The default only works for the default kernels.
    instrument_function : Callable, optional
         A function that takes data points (a list of dicts), and returns the same
         with the measurement data filled in. The function is
         expected to communicate with the instrument and perform measurements,
         populating fields of the data input. `y_data` and `noise variances` have to be filled in.
    init_dataset_size : int, optional
        If `x` and `y` are not provided and `dataset` is not provided,
        `init_dataset_size` must be provided. An initial
        dataset is constructed randomly with this length. The `instrument_function`
        is immediately called to measure values
        at these initial points.
    acquisition_function : Callable, optional
        The acquisition function accepts as input a numpy array
        of size V x D (such that V is the number of input
        points, and D is the parameter space dimensionality) and
        a `GPOptimizer` object. The return value is 1d array
        of length V providing 'scores' for each position,
        such that the highest scored point will be measured next.
        Built-in functions can be used by one of the following keys:
        `variance`, `relative information entropy`,
        `relative information entropy set`, `total correlation`.
        See fvGPOptimizer.ask() for a short explanation of these functions.
        In the multi-task case, it is highly recommended to
        deploy a user-defined acquisition function due to the intricate relationship
        of posterior distributions at different points in the output space.
        If None, the default function `variance`, meaning
        `fvgp.GP.posterior_covariance` with variance_only = True will be used.
        The acquisition function can be a callable function of the form my_func(x,gpcam.GPOptimizer)
        which will be maximized (!!!), so make sure desirable new measurement points
        will be located at maxima.
    cost_function : Callable, optional
        A function encoding the cost of motion through the input space and the
        cost of a measurements. Its inputs are an
        `origin` (np.ndarray of size V x D), `x` (np.ndarray of size V x D),
        and the value of `cost_function_parameters`;
        `origin` is the starting position, and `x` is the destination position. The return value is a 1d array of
        length V describing the costs as floats. The 'score' from
        acquisition_function is divided by this returned cost to determine
        the next measurement point. If None, the default is a uniform cost of 1.
    cost_update_function : Callable, optional
        A function that updates the `cost_func_params` which are communicated to the `cost_function`.
        This function accepts as input
        costs (a list of cost values determined by `instrument_function`), bounds (a V x 2 numpy array) and a parameters
        object. The default is a no-op.
    cost_function_parameters : Any, optional
        An object that is communicated to the `cost_function` and `cost_update_function`. The default is `{}`.
    online : bool, optional
        The default is True. `online=True` will lead to calls to `gpOptimizer.tell(append=True)` which
        potentially saves a lot of time in the GP update. The GP is updated either with an inversion update
        or a Cholesky factor update.
    kernel_function : Callable, optional
        A symmetric positive definite covariance function (a kernel)
        that calculates the covariance between
        data points. It is a function of the form k(x1,x2,hyperparameters).
        The input `x1` a N1 x Di+1 array of positions, `x2` is a N2 x Di+1
        array of positions, the hyperparameters argument
        is a 1d array of length N depending on how many hyperparameters are initialized.
        The default is a stationary anisotropic kernel
        (`fvgp.GP.default_kernel`) which performs automatic relevance determination (ARD). The task
        direction is simply considered an additional dimension. This kernel should only be used for tests and in the
        simplest of cases.
        The output is a matrix, an N1 x N2 numpy array.
    prior_mean_function : Callable, optional
        A function that evaluates the prior mean at a set of input position. It accepts as input
        an array of positions (of shape N1 x Di+1) and
        hyperparameters (a 1d array of length Di+2 for the default kernel).
        The return value is a 1d array of length N1. If None is provided,
        `fvgp.GP._default_mean_function` is used, which is the average of the `y_data`.
    noise_function : Callable optional
        The noise function is a callable f(x,hyperparameters) that returns a
        vector (1d np.ndarray) of length(x).
        The input `x` is a numpy array of shape (N x D). The hyperparameter array is the same
        that is communicated to mean and kernel functions.
        Only provide a noise function OR a noise variance vector, not both.
    run_every_iteration : Callable, optional
        A function that is run at every iteration. It accepts as input a
        `gpcam.AutonomousExperimenterGP` instance. The default is a no-op.
    x_data : np.ndarray, optional
        Initial data point positions.
    y_data : np.ndarray, optional
        Initial data point values.
    noise_variances : np.ndarray, optional
        Initial data point observation variances.
    dataset : string, optional
        A filename of a gpcam-generated file that is used to initialize a new instance.
    communicate_full_dataset : bool, optional
        If True, the full dataset will be communicated to the `instrument_function`
        on each iteration. If False, only the
        newly suggested data points will be communicated. The default is False.
    compute_device : str, optional
        One of `cpu` or `gpu`, determines how linear algebra computations are executed. The default is `cpu`.
    calc_inv : bool, optional
        If True, the algorithm calculates and stores the inverse of the covariance
        matrix after each training or update of the dataset or hyperparameters,
        which makes computing the posterior covariance faster (3-10 times).
        For larger problems (>2000 data points), the use of inversion should be avoided due
        to computational instability and costs. The default is
        False. Note, the training will not use the
        inverse for stability reasons. Storing the inverse is
        a good option when the dataset is not too large and the posterior covariance is heavily used.
        Caution: this option, together with `append=True` in `tell()` will mean that the inverse of
        the covariance is updated, not recomputed, which can lead to instability.
        In application where data is appended many times, it is recommended to either turn
        `calc_inv` off, or to regularly force the recomputation of the inverse via `gp_rank_n_update` in
        `update_gp_data`.
    training_dask_client : distributed.client.Client, optional
        A Dask Distributed Client instance for distributed training. If None is provided, a new
        `dask.distributed.Client` instance is constructed.
    acq_func_opt_dask_client : distributed.client.Client, optional
        A Dask Distributed Client instance for distributed `acquisition_function`
        optimization. If None is provided, a new
        `dask.distributed.Client` instance is constructed.


    Attributes
    ----------
    x_data : np.ndarray
        Data point positions
    y_data : np.ndarray
        Data point values
    noise_variances : np.ndarray
        Data point observation variances
    data.dataset : list
        All data
    hyperparameter_bounds : np.ndarray
        A 2d array of floats of size J x 2, such that J is the length
        matching the length of `hyperparameters` defining
        the bounds for training.
    gp_optimizer : gpcam.GPOptimizer
        A GPOptimizer instance used for initializing a Gaussian process and performing optimization of the posterior.
    """

    def __init__(self,
                 input_space,
                 hyperparameters=None,
                 hyperparameter_bounds=None,
                 instrument_function=None,
                 init_dataset_size=None,
                 acquisition_function="variance",
                 cost_function=None,
                 cost_update_function=None,
                 cost_function_parameters=None,
                 online=True,
                 kernel_function=None,
                 prior_mean_function=None,
                 noise_function=None,
                 run_every_iteration=None,
                 x_data=None, y_data=None, noise_variances=None, dataset=None,
                 communicate_full_dataset=False,
                 compute_device="cpu",
                 calc_inv=False,
                 training_dask_client=None,
                 acq_func_opt_dask_client=None,
                 gp2Scale=False,
                 gp2Scale_dask_client=None,
                 gp2Scale_batch_size=10000,
                 ram_economy=True,
                 args=None
                 ):
        ################################
        # getting the data ready#########
        ################################
        if cost_function_parameters is None:
            cost_function_parameters = {}
        dim = len(input_space)
        self.input_space = input_space
        self.instrument_function = instrument_function
        self.hyperparameter_bounds = hyperparameter_bounds
        self.hyperparameters = hyperparameters
        self.acquisition_function = acquisition_function
        self.run_every_iteration = run_every_iteration
        self.communicate_full_dataset = communicate_full_dataset
        self.async_train_in_progress = False
        self.training_dask_client = training_dask_client
        self.acq_func_opt_dask_client = acq_func_opt_dask_client
        self.args = args
        self.online = online

        if init_dataset_size is None and x_data is None and dataset is None:
            raise Exception("Either provide length of initial data or an initial dataset")
        self.data = fvgpData(dim, input_space)

        if x_data is None and dataset is None:
            self.data.create_random_dataset(init_dataset_size)
            if instrument_function is None: raise Exception("You need to provide an instrument function.")
            self.data.update_dataset(self.instrument_function(self.data.dataset))
        elif dataset is not None:
            self.data.inject_dataset(list(np.load(dataset, allow_pickle=True)))
            self.hyperparameters = self.data.dataset[-1]["hyperparameters"]
        elif x_data is not None and y_data is not None and noise_variances is not None:
            self.data.update_dataset(self.data.arrays2data(x_data, y=y_data, v=noise_variances))
        elif x_data is not None and y_data is None:
            if instrument_function is None: raise Exception("You need to provide an instrument function.")
            self.data.update_dataset(self.instrument_function(self.data.arrays2data(x_data,
                                                                                    y=y_data, v=noise_variances)))
        else:
            raise Exception("No viable option for data given!")
        self.data.check_incoming_data()
        self.x_data, self.y_data, self.noise_variances, self.times, self.costs = self.data.extract_data()
        self.init_dataset_size = len(self.x_data)
        self.multi_task = True
        ######################
        ######################
        ######################
        self.gp_optimizer = fvGPOptimizer(
            self.x_data, self.y_data,
            init_hyperparameters=hyperparameters,
            noise_variances=self.noise_variances,
            compute_device=compute_device,
            gp_kernel_function=kernel_function,
            gp_kernel_function_grad=None,
            gp_noise_function=noise_function,
            gp_noise_function_grad=None,
            gp_mean_function=prior_mean_function,
            gp_mean_function_grad=None,
            gp2Scale=gp2Scale,
            gp2Scale_dask_client=gp2Scale_dask_client,
            gp2Scale_batch_size=gp2Scale_batch_size,
            calc_inv=calc_inv,
            ram_economy=ram_economy,
            args=args,
            cost_function=cost_function,
            cost_function_parameters=cost_function_parameters,
            cost_update_function=cost_update_function)

        logger.debug(inspect.cleandoc("""#
        ##################################################################################
        Autonomous Experimenter fvGP initialization successfully concluded
        now train(...) or train_async(...), and then go(...)
        ##################################################################################"""))
